import inspect
from typing import Any, Callable, Dict, Optional

from .local import LocalPlaybook


class PythonPlaybook(LocalPlaybook):
    """Represents a Python playbook created from @playbook decorated functions.

    Python playbooks are defined using the @playbook decorator and contain
    executable Python code.
    """

    @classmethod
    def create_playbooks_from_h1(
        cls, h1: Dict, namespace_manager
    ) -> Dict[str, "PythonPlaybook"]:
        """Create PythonPlaybook instances from H1 AST node.

        Args:
            h1: H1 AST node containing agent definition
            namespace_manager: Namespace manager for executing code blocks

        Returns:
            Dict[str, PythonPlaybook]: Dictionary of created playbooks
        """
        playbooks = {}

        for child in h1.get("children", []):
            if child.get("type") == "code-block":
                new_playbooks = cls.create_playbooks_from_code_block(
                    child["text"],
                    namespace_manager,
                    child.get("source_file_path"),
                    child.get("line_number"),
                )
                playbooks.update(new_playbooks)

        return playbooks

    @classmethod
    def create_playbooks_from_code_block(
        cls,
        code_block: str,
        namespace_manager,
        source_file_path: Optional[str] = None,
        source_line_number: Optional[int] = None,
    ) -> Dict[str, "PythonPlaybook"]:
        """Create PythonPlaybook instances from a code block.

        Args:
            code_block: Python code containing @playbook decorated functions
            namespace_manager: Namespace manager for execution environment

        Returns:
            Dict[str, PythonPlaybook]: Dictionary of discovered playbooks
        """
        import ast

        # Set up the execution environment
        existing_keys = list(namespace_manager.namespace.keys())
        environment = namespace_manager.prepare_execution_environment()
        namespace_manager.namespace.update(environment)

        # Execute the code block in the isolated namespace
        python_local_namespace = {}
        exec(code_block, namespace_manager.namespace, python_local_namespace)
        namespace_manager.namespace.update(python_local_namespace)

        # Get code for each function
        function_code = {}
        items = {}
        parsed_code = ast.parse(code_block)
        # debug("Parsed Code: " + str(parsed_code))
        for item in parsed_code.body:
            if isinstance(item, ast.AsyncFunctionDef) or isinstance(
                item, ast.FunctionDef
            ):
                function_code[item.name] = ast.unparse(item)
                items[item.name] = item

        # Discover all @playbook-decorated functions
        playbooks = cls._discover_playbook_functions(namespace_manager, existing_keys)
        # debug("decorated functions: " + str(playbooks))
        # Add function code to playbooks
        for playbook in playbooks.values():
            playbook.code = function_code[playbook.name]
            playbook.source_file_path = source_file_path

            line_offset = 0
            if items[playbook.name].lineno is not None:
                line_offset = items[playbook.name].lineno
            playbook.source_line_number = source_line_number + line_offset

        return playbooks

    @classmethod
    def _discover_playbook_functions(
        cls, namespace_manager, existing_keys
    ) -> Dict[str, "PythonPlaybook"]:
        """Discover playbook-decorated functions in the namespace.

        Args:
            namespace_manager: Namespace manager containing executed code
            existing_keys: Keys that existed before code execution

        Returns:
            Dict[str, PythonPlaybook]: Discovered playbooks
        """
        import types

        playbooks = {}
        wrappers = {}

        for obj_name, obj in namespace_manager.namespace.items():
            if (
                isinstance(obj, types.FunctionType)
                and obj_name not in existing_keys
                and getattr(obj, "__is_playbook__", False)
            ):
                # Create playbook from decorated function
                playbooks[obj.__name__] = cls.from_function(obj)

                # Only create namespace functions if agent is available
                agent = namespace_manager.namespace.get("agent")
                if agent is not None:
                    wrappers[obj.__name__] = playbooks[
                        obj.__name__
                    ].create_namespace_function(agent)

        namespace_manager.namespace.update(wrappers)
        return playbooks

    @classmethod
    def from_function(cls, func: "Callable") -> "PythonPlaybook":
        """Create a PythonPlaybook object from a decorated function.

        Args:
            func: Decorated function

        Returns:
            PythonPlaybook: Created playbook object
        """
        import inspect
        import re

        from ..triggers import PlaybookTriggers

        sig = inspect.signature(func)
        signature = func.__name__ + str(sig)
        doc = inspect.getdoc(func)
        description = doc.split("\n")[0] if doc is not None else None
        triggers = getattr(func, "__triggers__", [])
        metadata = getattr(func, "__metadata__", {})

        # If triggers are not prefixed with T1:BGN, T1:CND, etc., add T{i}:CND
        # Use regex to find if prefix is missing
        triggers = [
            (
                f"T{i+1}:CND {trigger}"
                if not re.match(r"^T\d+:[A-Z]{3} ", trigger)
                else trigger
            )
            for i, trigger in enumerate(triggers)
        ]

        if triggers:
            triggers = PlaybookTriggers(
                playbook_klass=func.__name__,
                playbook_signature=signature,
                triggers=triggers,
            )
        else:
            triggers = None

        return cls(
            name=func.__name__,
            func=func,
            signature=signature,
            description=description,
            triggers=triggers,
            metadata=metadata,
        )

    def __init__(
        self,
        name: str,
        func: Callable,
        signature: str,
        description: Optional[str] = None,
        agent_name: Optional[str] = None,
        triggers: Optional[Any] = None,
        metadata: Optional[Dict[str, Any]] = None,
        code: Optional[str] = None,
        source_file_path: Optional[str] = None,
        source_line_number: Optional[int] = None,
    ):
        """Initialize a PythonPlaybook.

        Args:
            name: The name of the playbook (function name)
            func: The decorated function to execute
            signature: The function signature string
            description: Human-readable description of the playbook
            agent_name: Name of the agent this playbook belongs to
            triggers: Trigger configuration for the playbook
            metadata: Additional metadata for the playbook
            code: The source code of the function
            source_file_path: The file path of the source where this playbook is defined
            source_line_number: The line number in the source where this playbook is defined
        """
        super().__init__(
            name=name,
            description=description,
            agent_name=agent_name,
            metadata=metadata,
            source_file_path=source_file_path,
            source_line_number=source_line_number,
        )

        self.func = func
        self.signature = signature
        self.triggers = triggers
        self.code = code
        # For backward compatibility with existing code
        self.klass = name

    async def _execute_impl(self, *args, **kwargs) -> Any:
        """Execute the Python playbook function.

        Args:
            *args: Positional arguments for the playbook
            **kwargs: Keyword arguments for the playbook

        Returns:
            The result of executing the function
        """
        if not self.func:
            raise ValueError(f"PythonPlaybook {self.name} has no executable function")

        # Execute the function (it may be sync or async)
        if inspect.iscoroutinefunction(self.func):
            return await self.func(*args, **kwargs)
        else:
            return self.func(*args, **kwargs)

    def get_parameters(self) -> Dict[str, Any]:
        """Get the parameters schema for this playbook.

        Returns:
            A dictionary describing the expected parameters based on the function signature
        """
        if not self.func:
            return {}

        sig = inspect.signature(self.func)
        parameters = {}

        for param_name, param in sig.parameters.items():
            param_info = {
                "name": param_name,
                "kind": param.kind.name,
            }

            if param.annotation != inspect.Parameter.empty:
                param_info["type"] = str(param.annotation)

            if param.default != inspect.Parameter.empty:
                param_info["default"] = param.default

            parameters[param_name] = param_info

        return {
            "signature": self.signature,
            "parameters": parameters,
            "description": f"Parameters for {self.name} Python playbook",
        }

    def get_description(self) -> str:
        """Get a human-readable description of this playbook.

        Returns:
            The description of the playbook
        """
        return self.description or self.name

    def __repr__(self) -> str:
        """Return a string representation of the playbook."""
        return f"PythonPlaybook(name='{self.name}', agent='{self.agent_name}')"



"""Streaming wrapper for SessionLog that broadcasts entries via callbacks."""

import asyncio
from datetime import datetime
from typing import Callable, Optional

from playbooks.session_log import SessionLog, SessionLogItem, SessionLogItemLevel
from playbooks.session_log_items import SessionLogItemBase


class StreamingSessionLog(SessionLog):
    """
    A SessionLog wrapper that streams log entries via a callback.

    This allows web servers, UIs, or other consumers to receive
    real-time updates of session log entries.
    """

    def __init__(
        self, klass: str, agent_id: str, stream_callback: Optional[Callable] = None
    ):
        super().__init__(klass, agent_id)
        self.stream_callback = stream_callback

    def set_stream_callback(self, callback: Callable):
        """Set or update the stream callback and optionally stream existing entries."""
        self.stream_callback = callback

    def append(
        self,
        item: SessionLogItem | str,
        level: SessionLogItemLevel = SessionLogItemLevel.MEDIUM,
    ):
        """Override append to stream entries to callback."""
        # Call parent append first
        super().append(item, level)

        # Stream the entry if we have a callback
        if self.stream_callback and self.log:
            # Get the last appended entry
            last_entry = self.log[-1]
            item = last_entry["item"]

            # Create streaming event data
            event_data = {
                "timestamp": datetime.now().isoformat(),
                "agent_id": self.agent_id,
                "agent_klass": self.klass,
                "level": level.name,
                "content": str(item),
            }

            # Add metadata if it's an enhanced SessionLogItem
            if isinstance(item, SessionLogItemBase):
                event_data["metadata"] = item.to_metadata()
                event_data["item_type"] = item.item_type
            elif hasattr(item, "__class__"):
                # For existing item types like PlaybookCall
                event_data["item_type"] = item.__class__.__name__.lower()
            else:
                event_data["item_type"] = "message"

            # Add different log representations
            if hasattr(item, "to_log_full"):
                event_data["log_full"] = item.to_log_full()
            if hasattr(item, "to_log_compact"):
                event_data["log_compact"] = item.to_log_compact()
            if hasattr(item, "to_log_minimal"):
                event_data["log_minimal"] = item.to_log_minimal()

            # Call the callback
            if asyncio.iscoroutinefunction(self.stream_callback):
                # If callback is async, create a task to run it
                asyncio.create_task(self.stream_callback(event_data))
            else:
                # If callback is sync, just call it
                self.stream_callback(event_data)

    @classmethod
    def wrap_existing(
        cls, existing_log: SessionLog, stream_callback: Optional[Callable] = None
    ) -> "StreamingSessionLog":
        """
        Wrap an existing SessionLog to add streaming capability.

        This is useful when you want to retrofit streaming onto
        an already-created SessionLog.
        """
        streaming_log = cls(existing_log.klass, existing_log.agent_id, stream_callback)
        streaming_log.log = existing_log.log
        return streaming_log



"""
Minimal logging configuration for Playbooks framework.

Environment Variables:
    PLAYBOOKS_DEBUG=false           # Enable/disable debug output
    PLAYBOOKS_OUTPUT_HANDLER=console # console|websocket|silent
    PLAYBOOKS_DEBUG_FILE=           # Optional debug log file path
"""

import os
from typing import Optional

from .logging_constants import (
    ENV_DEBUG_ENABLED,
    ENV_OUTPUT_HANDLER,
    ENV_DEBUG_FILE,
    DEFAULT_DEBUG_ENABLED,
    DEFAULT_OUTPUT_HANDLER,
    HANDLER_CONSOLE,
    HANDLER_SILENT,
    parse_boolean_env,
)


class Config:
    """Minimal logging configuration with sensible defaults."""

    DEBUG_ENABLED: bool = parse_boolean_env(
        os.getenv(ENV_DEBUG_ENABLED, DEFAULT_DEBUG_ENABLED)
    )
    OUTPUT_HANDLER: str = os.getenv(ENV_OUTPUT_HANDLER, DEFAULT_OUTPUT_HANDLER)
    DEBUG_FILE: Optional[str] = os.getenv(ENV_DEBUG_FILE)

    @classmethod
    def is_debug_enabled(cls) -> bool:
        """Check if debug logging is enabled."""
        return cls.DEBUG_ENABLED

    @classmethod
    def get_output_handler(cls) -> str:
        """Get the configured output handler type."""
        return cls.OUTPUT_HANDLER

    @classmethod
    def get_debug_file(cls) -> Optional[str]:
        """Get the debug file path if configured."""
        return cls.DEBUG_FILE


def apply_config() -> None:
    """Apply configuration to logging system."""
    from playbooks.user_output import setup_for_console, setup_for_testing

    handler_type = Config.get_output_handler()

    if handler_type == HANDLER_CONSOLE:
        setup_for_console()
    elif handler_type == HANDLER_SILENT:
        setup_for_testing()
    # WebSocket handler is set up explicitly in web applications
    # via setup_for_web(emit_function)



from typing import Any, Dict, List, Optional, Union

from .call_stack import InstructionPointer
from .event_bus import EventBus
from .events import VariableUpdateEvent


class VariableChangeHistoryEntry:
    def __init__(self, instruction_pointer: InstructionPointer, value: Any):
        self.instruction_pointer = instruction_pointer
        self.value = value


class Variable:
    def __init__(self, name: str, value: Any):
        self.name = name
        self.value = value
        self.change_history: List[VariableChangeHistoryEntry] = []

    def update(
        self, new_value: Any, instruction_pointer: Optional[InstructionPointer] = None
    ):
        self.change_history.append(
            VariableChangeHistoryEntry(instruction_pointer, new_value)
        )
        self.value = new_value

    def __repr__(self) -> str:
        return f"{self.name}={self.value}"


class Variables:
    """A collection of variables with change history."""

    def __init__(self, event_bus: EventBus, agent_id: str = "unknown"):
        self.variables: Dict[str, Variable] = {}
        self.event_bus = event_bus
        self.agent_id = agent_id

    def update(self, vars: Union["Variables", Dict[str, Any]]) -> None:
        """Update multiple variables at once."""
        if isinstance(vars, Variables):
            for name, value in vars.variables.items():
                self[name] = value.value
        else:
            for name, value in vars.items():
                self[name] = value

    def __getitem__(self, name: str) -> Variable:
        return self.variables[name]

    def __setitem__(
        self,
        name: str,
        value: Any,
        instruction_pointer: Optional[InstructionPointer] = None,
    ) -> None:
        if ":" in name:
            name = name.split(":")[0]
        if name not in self.variables:
            self.variables[name] = Variable(name, value)
        self.variables[name].update(value, instruction_pointer)
        event = VariableUpdateEvent(
            agent_id=self.agent_id,
            session_id="",
            variable_name=name,
            variable_value=value,
        )
        self.event_bus.publish(event)

    def __contains__(self, name: str) -> bool:
        return name in self.variables

    def __iter__(self):
        return iter(self.variables.values())

    def __len__(self) -> int:
        return len(self.variables)

    def to_dict(self, include_private: bool = False) -> Dict[str, Any]:
        return {
            name: variable.value
            for name, variable in self.variables.items()
            if variable.value is not None
            and (include_private or not variable.name.startswith("$_"))
        }

    def __repr__(self) -> str:
        return f"Variables({self.to_dict(include_private=True)})"



from glob import glob
from pathlib import Path
from typing import List, Tuple

from .exceptions import ProgramLoadError
from .import_processor import ImportProcessor
from .utils.file_utils import is_compiled_playbook_file


class Loader:
    @staticmethod
    def read_program(program_paths: List[str]) -> str:
        """
        Load program content from file paths.
        """
        program_content = None
        try:
            program_content, do_not_compile = Loader._read_program(program_paths)
        except FileNotFoundError as e:
            raise ProgramLoadError(str(e)) from e
        except (OSError, IOError) as e:
            raise ProgramLoadError(str(e)) from e

        return program_content, do_not_compile

    @staticmethod
    def _read_program(paths: List[str]) -> str:
        """
        Load program content from file paths. Supports both single files and glob patterns.

        Args:
            paths: List of file paths or glob patterns (e.g., 'my_playbooks/**/*.pb')

        Returns:
            str: Combined contents of all matching program files

        Raises:
            FileNotFoundError: If no files are found or if files are empty
        """
        all_files = []

        for path in paths:
            # Simplified glob pattern check
            if "*" in str(path) or "?" in str(path) or "[" in str(path):
                # Handle glob pattern
                all_files.extend(glob(path, recursive=True))
            else:
                # Handle single file
                all_files.append(path)

        if not all_files:
            raise FileNotFoundError("No files found")

        # Deduplicate files and read content
        contents = []
        do_not_compile = False
        not_found = []
        for file in set(all_files):
            file_path = Path(file)
            if file_path.is_file() and file_path.exists():
                if is_compiled_playbook_file(file_path):
                    do_not_compile = True
                contents.append(file_path.read_text())
            else:
                not_found.append(str(file_path))

        if not_found:
            raise FileNotFoundError(f"{', '.join(not_found)} not found")

        program_contents = "\n\n".join(contents)

        if not program_contents:
            raise FileNotFoundError("Files found but content is empty")

        return program_contents, do_not_compile

    @staticmethod
    def read_program_files(program_paths: List[str]) -> List[Tuple[str, str, bool]]:
        """
        Load program files individually.

        Args:
            program_paths: List of file paths or glob patterns

        Returns:
            List of (file_path, content, is_compiled) tuples

        Raises:
            ProgramLoadError: If files cannot be read
        """
        try:
            return Loader._read_program_files(program_paths)
        except FileNotFoundError as e:
            raise ProgramLoadError(str(e)) from e
        except (OSError, IOError) as e:
            raise ProgramLoadError(str(e)) from e

    @staticmethod
    def _read_program_files(paths: List[str]) -> List[Tuple[str, str, bool]]:
        """
        Load program files individually with their metadata.
        Processes !import directives in non-compiled files.

        Args:
            paths: List of file paths or glob patterns

        Returns:
            List of (file_path, content, is_compiled) tuples

        Raises:
            FileNotFoundError: If no files are found or cannot be read
        """
        all_files = []

        for path in paths:
            # Simplified glob pattern check
            if "*" in str(path) or "?" in str(path) or "[" in str(path):
                # Handle glob pattern
                all_files.extend(glob(path, recursive=True))
            else:
                # Handle single file
                all_files.append(path)

        if not all_files:
            raise FileNotFoundError("No files found")

        # Initialize import processor
        import_processor = ImportProcessor()

        # Read files individually
        files_data = []
        not_found = []

        for file in all_files:
            file_path = Path(file)
            if file_path.is_file() and file_path.exists():
                content = file_path.read_text()
                is_compiled = is_compiled_playbook_file(file_path)

                # Process imports for non-compiled files
                if not is_compiled and "!import" in content:
                    try:
                        # Reset processor for each file to avoid cross-file state
                        import_processor.reset()
                        content = import_processor.process_imports(content, file_path)
                    except ProgramLoadError as e:
                        # Re-raise with more context
                        raise ProgramLoadError(
                            f"Error processing imports in {file_path}: {str(e)}"
                        ) from e

                files_data.append((str(file_path), content, is_compiled))
            else:
                not_found.append(str(file_path))

        if not_found:
            raise FileNotFoundError(f"{', '.join(not_found)} not found")

        if not files_data:
            raise FileNotFoundError("No valid files found")

        # Check for empty content
        if all(not content.strip() for _, content, _ in files_data):
            raise FileNotFoundError("Files found but all content is empty")

        return files_data



"""Base LLMMessage class for handling LLM messages."""

from typing import Any, Dict, Optional

from playbooks.enums import LLMMessageRole, LLMMessageType


class LLMMessage:
    """Base class for all LLM messages.

    This class provides a structured way to handle messages sent to and from LLMs,
    with support for different roles, types, and caching strategies.

    Attributes:
        content: The text content of the message
        role: The role of the message sender (system, user, assistant)
        type: The type of message
        cached: Whether this message should be cached by the LLM provider
    """

    # Class-level constants for validation
    MAX_CONTENT_SIZE = 100_000  # 100K char limit
    MIN_CONTENT_SIZE = 0  # Allow empty content for some use cases

    def __init__(
        self,
        content: str,
        role: LLMMessageRole,
        type: LLMMessageType = LLMMessageType.USER_INPUT,
        cached: bool = False,
    ) -> None:
        """Initialize an LLMMessage.

        Args:
            content: The text content of the message
            role: The role of the message sender
            type: The type of message (defaults to USER_INPUT)
            cached: Whether this message should be cached

        Raises:
            TypeError: If content is not a string
            ValueError: If content or other parameters are invalid
        """
        # Validate and normalize content
        self._content = self._validate_content(content)

        # Validate role and type
        self._role = self._validate_role(role)
        self._type = self._validate_type(type)

        # Validate cached flag
        self._cached = self._validate_cached(cached)

    @staticmethod
    def _validate_content(content: str) -> str:
        """Validate and normalize content.

        Args:
            content: The content to validate

        Returns:
            The validated content string

        Raises:
            TypeError: If content is not a string
            ValueError: If content is invalid
        """
        if content is None:
            raise TypeError("content cannot be None")

        if not isinstance(content, str):
            raise TypeError(f"content must be a string, got {type(content).__name__}")

        # Check size limits
        if len(content) > LLMMessage.MAX_CONTENT_SIZE:
            raise ValueError(
                f"content too large: {len(content)} bytes > {LLMMessage.MAX_CONTENT_SIZE} bytes"
            )

        # Content can be empty for some valid use cases, so we allow it
        # But we normalize whitespace-only strings to empty strings
        normalized = content.strip()
        if not normalized and content:  # Had content but only whitespace
            # For most cases, whitespace-only content is probably an error
            # but we'll allow it and normalize to empty string
            return ""

        return content  # Return original to preserve intentional whitespace

    @staticmethod
    def _validate_role(role: LLMMessageRole) -> LLMMessageRole:
        """Validate role parameter.

        Args:
            role: The role to validate

        Returns:
            The validated role

        Raises:
            TypeError: If role is not an LLMMessageRole
        """
        if not isinstance(role, LLMMessageRole):
            raise TypeError(
                f"role must be an LLMMessageRole, got {type(role).__name__}"
            )
        return role

    @staticmethod
    def _validate_type(msg_type: LLMMessageType) -> LLMMessageType:
        """Validate type parameter.

        Args:
            msg_type: The message type to validate

        Returns:
            The validated message type

        Raises:
            TypeError: If msg_type is not an LLMMessageType
        """
        if not isinstance(msg_type, LLMMessageType):
            raise TypeError(
                f"type must be an LLMMessageType, got {type(msg_type).__name__}"
            )
        return msg_type

    @staticmethod
    def _validate_cached(cached: bool) -> bool:
        """Validate cached parameter.

        Args:
            cached: The cached flag to validate

        Returns:
            The validated cached flag

        Raises:
            TypeError: If cached is not a boolean
        """
        if not isinstance(cached, bool):
            raise TypeError(f"cached must be a boolean, got {type(cached).__name__}")
        return cached

    @staticmethod
    def _validate_string_param(
        value: Any, param_name: str, allow_empty: bool = False
    ) -> str:
        """Validate a string parameter with common rules.

        Args:
            value: The value to validate
            param_name: Name of the parameter for error messages
            allow_empty: Whether to allow empty strings

        Returns:
            The validated string

        Raises:
            TypeError: If value is not a string
            ValueError: If value is invalid
        """
        if value is None:
            raise TypeError(f"{param_name} cannot be None")

        if not isinstance(value, str):
            raise TypeError(
                f"{param_name} must be a string, got {type(value).__name__}"
            )

        # Check for whitespace-only strings
        if not value.strip():
            if allow_empty and not value:
                return value  # Allow truly empty strings if specified
            raise ValueError(f"{param_name} cannot be empty or whitespace-only")

        return value

    # Properties for immutability
    @property
    def content(self) -> str:
        """Get the message content."""
        return self._content

    @property
    def role(self) -> LLMMessageRole:
        """Get the message role."""
        return self._role

    @property
    def type(self) -> LLMMessageType:
        """Get the message type."""
        return self._type

    @property
    def cached(self) -> bool:
        """Get the cached flag."""
        return self._cached

    def to_full_message(self) -> Dict[str, Any]:
        """Convert to full dictionary representation for LLM APIs.

        Returns:
            A dictionary with role, type, content, and optionally cache_control fields
        """
        message = {
            "role": self.role.value,
            "type": self.type.value,
            "content": self.content,
        }
        if self.cached:
            message["cache_control"] = {"type": "ephemeral"}
        return message

    def to_compact_message(self) -> Optional[Dict[str, Any]]:
        """Convert to compact representation for token optimization.

        This method can be extended in the future to provide
        progressive compaction strategies for reducing token usage.

        Returns:
            A dictionary representation or None to remove completely
        """
        # Default implementation returns full message
        # Can be overridden in subclasses for specific compaction strategies
        return self.to_full_message()

    def __repr__(self) -> str:
        """String representation of the message."""
        cached_str = ", cached=True" if self.cached else ""
        return f"{self.__class__.__name__}(role={self.role}, type={self.type}, content_length={len(self.content)}{cached_str})"

    def __eq__(self, other: object) -> bool:
        """Check equality with another message.

        Subclasses should override this to include their custom attributes.
        """
        if not isinstance(other, self.__class__):
            return False
        return (
            self.content == other.content
            and self.role == other.role
            and self.type == other.type
            and self.cached == other.cached
        )

    def __hash__(self) -> int:
        """Make messages hashable for use in sets and as dict keys.

        Subclasses should override this to include their custom attributes.
        """
        return hash(
            (self.__class__.__name__, self.content, self.role, self.type, self.cached)
        )



"""Execution state management for the interpreter.

This module provides the ExecutionState class, which encapsulates the state
tracked during interpreter execution, including call stack, exit conditions,
and execution control flags.
"""

from typing import Any, Dict, List, Optional

from playbooks.artifacts import Artifacts
from playbooks.call_stack import CallStack
from playbooks.event_bus import EventBus
from playbooks.meetings import JoinedMeeting, Meeting
from playbooks.session_log import SessionLog
from playbooks.variables import Variables


class ExecutionState:
    """Encapsulates execution state including call stack, variables, and artifacts.

    Attributes:
        bus: The event bus
        session_log: Log of session activity
        call_stack: Stack tracking the execution path
        variables: Collection of variables with change history
        artifacts: Store for execution artifacts
    """

    def __init__(self, event_bus: EventBus, klass: str, agent_id: str):
        """Initialize execution state with an event bus.

        Args:
            bus: The event bus to use for all components
        """
        self.event_bus = event_bus
        self.klass = klass
        self.agent_id = agent_id
        self.session_log = SessionLog(klass, agent_id)
        self.call_stack = CallStack(event_bus, agent_id)
        self.variables = Variables(event_bus, agent_id)
        self.artifacts = Artifacts()
        self.agents: List[Dict[str, Any]] = []
        self.last_llm_response = ""
        self.last_message_target = (
            None  # Track last 1:1 message target for Say() fallback
        )

        # Meetings initiated by this agent (agent is the owner/host)
        self.owned_meetings: Dict[str, "Meeting"] = {}  # meeting_id -> Meeting

        # Meetings this agent has joined as a participant
        self.joined_meetings: Dict[str, JoinedMeeting] = {}

    def __repr__(self) -> str:
        """Return a string representation of the execution state."""
        return f"{self.call_stack.__repr__()};{self.variables.__repr__()};{self.artifacts.__repr__()}"

    def to_dict(self) -> Dict[str, Any]:
        """Return a dictionary representation of the execution state."""
        # Build meetings list for LLM visibility from owned meetings
        meetings_list = []
        for meeting_id, meeting in self.owned_meetings.items():
            participants_list = []
            for participant in meeting.joined_attendees:
                participants_list.append(f"{participant.klass}(agent {participant.id})")

            meetings_list.append(
                {"meeting_id": meeting_id, "participants": participants_list}
            )

        return {
            "call_stack": [
                frame.instruction_pointer.to_compact_str()
                for frame in self.call_stack.frames
            ],
            "variables": self.variables.to_dict(),
            "artifacts": self.artifacts.to_dict(),
            "agents": self.agents,
            "meetings": meetings_list,
        }

    def __str__(self) -> str:
        """Return a string representation of the execution state."""
        return f"ExecutionState(call_stack={self.call_stack}, variables={self.variables}, session_log={self.session_log})"

    def get_current_meeting(self) -> Optional[str]:
        """Get meeting ID from top meeting playbook in call stack.

        Returns:
            Meeting ID if currently in a meeting, None otherwise
        """
        for frame in reversed(self.call_stack.frames):
            if frame.is_meeting and frame.meeting_id:
                return frame.meeting_id
        return None



import os
from typing import Any

from langfuse import Langfuse

from ..config import config


class PlaybooksLangfuseSpan:
    def update(self, **kwargs: Any) -> None:
        pass

    def generation(self, **kwargs: Any) -> None:
        pass

    def span(self, **kwargs: Any) -> None:
        return PlaybooksLangfuseSpan()


class PlaybooksLangfuseInstance:
    def trace(self, **kwargs: Any) -> None:
        return PlaybooksLangfuseSpan()

    def flush(self) -> None:
        pass


class LangfuseHelper:
    """A singleton helper class for Langfuse telemetry and tracing.

    This class provides centralized access to Langfuse for observability and
    tracing of LLM operations throughout the application.
    """

    langfuse: Langfuse | None = None

    @classmethod
    def instance(cls) -> Langfuse | None:
        """Get or initialize the Langfuse singleton instance.

        Creates the Langfuse client on first call using environment variables.
        Returns None if environment variables are not set.

        Returns:
            Langfuse: The initialized Langfuse client instance, or None
        """
        if cls.langfuse is None:
            # Check if Langfuse is enabled via config system first, then env fallback
            langfuse_enabled = False
            try:
                langfuse_enabled = config.langfuse.enabled
            except Exception:
                # Fallback to environment variable if config loading fails
                langfuse_enabled = (
                    os.getenv("LANGFUSE_ENABLED", "false").lower() == "true"
                )

            if not langfuse_enabled:
                cls.langfuse = PlaybooksLangfuseInstance()
            else:
                cls.langfuse = Langfuse(
                    secret_key=os.getenv("LANGFUSE_SECRET_KEY"),
                    public_key=os.getenv("LANGFUSE_PUBLIC_KEY"),
                    host=os.getenv("LANGFUSE_HOST"),
                )
        return cls.langfuse

    @classmethod
    def flush(cls) -> None:
        """Flush any buffered Langfuse telemetry data to the server.

        This method should be called when immediate data transmission is needed,
        such as before application shutdown or after important operations.
        """
        if cls.langfuse:
            cls.langfuse.flush()



from abc import ABC, ABCMeta
from typing import TYPE_CHECKING, Any, Dict

from playbooks.debug_logger import debug
from playbooks.events import AgentPausedEvent
from playbooks.utils.spec_utils import SpecUtils

from ..llm_messages import AgentCommunicationLLMMessage
from .messaging_mixin import MessagingMixin

if TYPE_CHECKING:
    from src.playbooks.program import Program


class BaseAgentMeta(ABCMeta):
    """Meta class for BaseAgent."""

    def should_create_instance_at_start(self) -> bool:
        """Whether to create an instance of the agent at start.

        Override in subclasses to control whether to create an instance at start.
        """
        return False


class BaseAgent(MessagingMixin, ABC, metaclass=BaseAgentMeta):
    """
    Base class for all agent implementations.

    Agents define behavior - what they do, their methods, and internal state.
    The runtime (Program) decides when and where they run.
    """

    def __init__(
        self,
        agent_id: str,
        program: "Program",
        source_line_number: int = None,
        source_file_path: str = None,
        **kwargs,
    ):
        """Initialize a new BaseAgent."""
        super().__init__()
        self.klass = self.__class__.klass
        self.description = self.__class__.description
        self.metadata = self.__class__.metadata.copy()

        self.id = agent_id
        self.kwargs = kwargs
        self.program = program

        # Source tracking
        self.source_line_number = source_line_number
        self.source_file_path = source_file_path

        # Debug context
        self._debug_thread_id: int = None
        self.paused: str = None

    async def begin(self):
        """Agent startup logic. Override in subclasses."""
        pass

    async def initialize(self):
        """Agent initialization logic. Override in subclasses."""
        pass

    # Built-in playbook methods
    async def Say(self, target: str, message: str, already_streamed: bool = False):
        resolved_target = self.resolve_target(target, allow_fallback=True)

        # Handle meeting targets with broadcasting
        if SpecUtils.is_meeting_spec(resolved_target):
            meeting_id = SpecUtils.extract_meeting_id(resolved_target)
            if (
                hasattr(self, "state")
                and hasattr(self.state, "owned_meetings")
                and meeting_id in self.state.owned_meetings
            ):
                debug(
                    f"{str(self)}: Broadcasting to meeting {meeting_id} as owner: {message}"
                )
                await self.meeting_manager.broadcast_to_meeting_as_owner(
                    meeting_id, message
                )
            elif (
                hasattr(self, "state")
                and hasattr(self.state, "joined_meetings")
                and meeting_id in self.state.joined_meetings
            ):
                debug(
                    f"{str(self)}: Broadcasting to meeting {meeting_id} as participant: {message}"
                )
                await self.meeting_manager.broadcast_to_meeting_as_participant(
                    meeting_id, message
                )
            else:
                # Error: not in this meeting
                debug(f"{str(self)}: state {self.state.joined_meetings}")
                debug(
                    f"{str(self)}: Cannot broadcast to meeting {meeting_id} - not a participant"
                )
                self.state.session_log.append(
                    f"Cannot broadcast to meeting {meeting_id} - not a participant"
                )
            return

        # Track last message target (only for 1:1 messages, not meetings)
        if not (
            SpecUtils.is_meeting_spec(resolved_target) or resolved_target == "human"
        ):
            self.state.last_message_target = resolved_target

        if not already_streamed and resolved_target == "human":
            await self.start_streaming_say()
            await self.stream_say_update(message)
            await self.complete_streaming_say()

        await self.SendMessage(resolved_target, message)

    async def SendMessage(self, target_agent_id: str, message: str):
        """Send a message to another agent."""
        if not self.program:
            return

        # Add to current frame context if available
        if hasattr(self, "state") and self.state.call_stack.peek():
            current_frame = self.state.call_stack.peek()
            if current_frame.playbook == "Say":
                current_frame = self.state.call_stack.frames[-2]
            target_agent = self.program.agents_by_id.get(target_agent_id)
            target_name = (
                str(target_agent)
                if target_agent
                else self.unknown_agent_str(target_agent_id)
            )
            agent_comm_msg = AgentCommunicationLLMMessage(
                f"I {str(self)} sent message to {target_name}: {message}",
                sender_agent=self.klass,
                target_agent=target_name,
            )
            current_frame.add_llm_message(agent_comm_msg)

        # Route through program runtime
        await self.program.route_message(
            sender_id=self.id,
            sender_klass=self.klass,
            receiver_spec=target_agent_id,
            message=message,
        )

    async def start_streaming_say(self, recipient=None):
        """Start displaying a streaming Say() message. Override in subclasses."""
        pass

    async def stream_say_update(self, content: str):
        """Add content to the current streaming Say() message. Override in subclasses."""
        pass

    async def complete_streaming_say(self):
        """Complete the current streaming Say() message. Override in subclasses."""
        pass

    def to_dict(self) -> Dict[str, Any]:
        return {**self.kwargs, "type": self.klass, "agent_id": self.id}

    def get_debug_thread_id(self) -> int:
        """Get the debug thread ID for this agent."""
        return self._debug_thread_id

    def set_debug_thread_id(self, thread_id: int) -> None:
        """Set the debug thread ID for this agent."""
        self._debug_thread_id = thread_id

    def emit_agent_paused_event(
        self, reason: str = "pause", source_line_number: int = 0
    ) -> None:
        """Emit an agent paused event for debugging."""
        if (
            self.program
            and hasattr(self.program, "event_bus")
            and self.program.event_bus
        ):
            event = AgentPausedEvent(
                session_id="",
                agent_id=self.id,
                reason=reason,
                source_line_number=source_line_number,
            )
            self.program.event_bus.publish(event)

    def emit_agent_resumed_event(self) -> None:
        """Emit an agent resumed event for debugging."""
        if (
            self.program
            and hasattr(self.program, "event_bus")
            and self.program.event_bus
        ):
            from playbooks.events import AgentResumedEvent

            event = AgentResumedEvent(
                session_id="",
                agent_id=self.id,
            )
            self.program.event_bus.publish(event)



import re
from typing import TYPE_CHECKING, Dict, Optional, Type, Union

from playbooks.exceptions import AgentConfigurationError
from playbooks.utils.markdown_to_ast import refresh_markdown_attributes
from playbooks.utils.parse_utils import parse_metadata_and_description
from playbooks.utils.text_utils import is_camel_case, to_camel_case

from . import LocalAIAgent, MCPAgent
from .builtin_playbooks import BuiltinPlaybooks
from .namespace_manager import AgentNamespaceManager
from .registry import AgentClassRegistry

if TYPE_CHECKING:
    pass


class AgentBuilder:
    """
    This class creates Agent classes based on the Abstract Syntax Tree
    representation of playbooks.
    """

    def __init__(self):
        """Initialize a new AgentBuilder instance."""
        self.namespace_manager = AgentNamespaceManager()
        self.builtin_playbooks = BuiltinPlaybooks()
        self.playbooks = {}

    @classmethod
    def create_agent_classes_from_ast(
        cls, ast: Dict
    ) -> Dict[str, Type[Union[LocalAIAgent, MCPAgent]]]:
        """
        Create agent classes from the AST representation of playbooks.

        Args:
            ast: AST dictionary containing playbook definitions

        Returns:
            Dict[str, Type[Union[LocalAIAgent, MCPAgent]]]: Dictionary mapping agent names to their classes
        """
        agents = {}
        for h1 in ast.get("children", []):
            if h1.get("type") == "h1":
                agent_name = h1["text"].strip()

                if not is_camel_case(agent_name):
                    agent_name = to_camel_case(agent_name)
                    h1["text"] = agent_name
                    refresh_markdown_attributes(h1)

                builder = cls()
                h1["children"].extend(builder.builtin_playbooks.get_ast_nodes())
                agents[agent_name] = builder.create_agent_class_from_h1(h1)

        return agents

    def create_agent_class_from_h1(
        self, h1: Dict
    ) -> Type[Union[LocalAIAgent, MCPAgent]]:
        """
        Create an Agent class from an H1 section in the AST.

        Args:
            h1: Dictionary representing an H1 section from the AST

        Returns:
            Type[Union[LocalAIAgent, MCPAgent]]: Dynamically created Agent class

        Raises:
            AgentConfigurationError: If agent configuration is invalid
        """
        klass = h1["text"].strip()

        # Check if agent name is provided
        if not klass:
            raise AgentConfigurationError("Agent name is required")

        # Check if class name is a valid CamelCase class name
        if not self.check_camelcase(klass):
            raise AgentConfigurationError(
                f"Agent name '{klass}' is not a valid CamelCase class name"
            )

        # Check if class already exists
        if klass in globals():
            raise AgentConfigurationError(f"Duplicate agent class {klass}")

        description = self._extract_description(h1)

        # Parse metadata to check for remote configuration
        metadata, description = parse_metadata_and_description(description)

        # Find appropriate agent class using registry
        agent_class = AgentClassRegistry.find_agent_class(metadata)
        if agent_class is None:
            raise AgentConfigurationError(
                f"No agent class can handle the configuration for agent {klass}. "
                f"Metadata: {metadata}"
            )

        # Create agent class using the found agent class factory
        return agent_class.create_class(
            klass,
            description,
            metadata,
            h1,
            h1.get("line_number"),
            self.namespace_manager,
        )

    @staticmethod
    def _extract_description(h1: Dict) -> Optional[str]:
        """
        Extract description from H1 node.

        Args:
            h1: Dictionary representing an H1 section from the AST

        Returns:
            Optional[str]: description or None if no description
        """
        description_parts = []

        for child in h1.get("children", []):
            if child.get("type") == "paragraph" or child.get("type") == "hr":
                description_text = child.get("text", "").strip()
                if description_text:
                    description_parts.append(description_text)

        description = "\n".join(description_parts).strip() or None
        return description

    # @staticmethod
    # def make_agent_class_name(klass: str) -> str:
    #     """Convert a string to a valid CamelCase class name prefixed with "Agent".

    #     Args:
    #         klass: Input string to convert to class name

    #     Returns:
    #         str: CamelCase class name prefixed with "Agent"

    #     Example:
    #         Input:  "This    is my agent!"
    #         Output: "AgentThisIsMyAgent"
    #     """
    #     import re

    #     # Replace any non-alphanumeric characters with a single space
    #     cleaned = re.sub(r"[^A-Za-z0-9]+", " ", klass)

    #     # Split on whitespace and filter out empty strings
    #     words = [w for w in cleaned.split() if w]

    #     # Capitalize each word and join
    #     capitalized_words = [w.capitalize() for w in words]

    #     # Prefix with "Agent" and return
    #     return "Agent" + "".join(capitalized_words)

    def check_camelcase(self, str: str) -> bool:
        """Check if a string is a valid CamelCase class name."""
        # Allow standard PascalCase with letters and numbers
        pattern = "^[A-Z][a-zA-Z0-9]*$"
        if re.match(pattern, str):
            return True
        else:
            return False



from abc import ABC, abstractmethod
from typing import Any, Callable, Dict, List, Optional

from ..triggers import PlaybookTriggers


class Playbook(ABC):
    """Abstract base class for all playbook implementations.

    This class defines the interface that all playbook types must implement,
    whether they are local markdown playbooks, remote MCP playbooks, or other types.
    """

    def __init__(
        self,
        name: str,
        description: Optional[str] = None,
        agent_name: Optional[str] = None,
        metadata: Optional[Dict[str, Any]] = None,
    ):
        """Initialize a playbook.

        Args:
            name: The name/class of the playbook
            description: Human-readable description of the playbook
            agent_name: Name of the agent this playbook belongs to
            metadata: Additional metadata for the playbook
        """
        self.name = name
        self.description = description
        self.resolved_description = description
        self.agent_name = agent_name
        self.metadata = metadata or {}
        self.triggers: Optional[PlaybookTriggers] = None

    @property
    def public(self) -> bool:
        """Return whether this playbook is public.

        Public playbooks can be called by other agents.

        Returns:
            True if the playbook is public, False otherwise
        """
        return self.metadata.get("public", False)

    @property
    def export(self) -> bool:
        """Return whether this playbook is exported.

        Exported playbooks are available for external use.

        Returns:
            True if the playbook is exported, False otherwise
        """
        return self.metadata.get("export", False)

    @property
    def hidden(self) -> bool:
        """Return whether this playbook is hidden.

        Hidden playbooks are not shown in the agent's public information.
        """
        return self.metadata.get("hidden", False)

    @property
    def meeting(self) -> bool:
        """Return whether this playbook is a meeting playbook.

        Meeting playbooks are designed to orchestrate meetings with multiple participants.

        Returns:
            True if the playbook is a meeting playbook, False otherwise
        """
        return self.metadata.get("meeting", False)

    @property
    def required_attendees(self) -> List[str]:
        """Return the list of required attendees for meeting playbooks.

        Returns:
            List of required attendee identifiers, empty list if none specified
        """
        return self.metadata.get("required_attendees", [])

    @property
    def optional_attendees(self) -> List[str]:
        """Return the list of optional attendees for meeting playbooks.

        Returns:
            List of optional attendee identifiers, empty list if none specified
        """
        return self.metadata.get("optional_attendees", [])

    @abstractmethod
    async def execute(self, *args, **kwargs) -> Any:
        """Execute the playbook with the given arguments.

        Args:
            *args: Positional arguments for the playbook
            **kwargs: Keyword arguments for the playbook

        Returns:
            The result of executing the playbook

        Raises:
            Exception: If execution fails
        """
        pass

    @abstractmethod
    def get_parameters(self) -> Dict[str, Any]:
        """Get the parameters schema for this playbook.

        Returns:
            A dictionary describing the expected parameters
        """
        pass

    @abstractmethod
    def get_description(self) -> str:
        """Get a human-readable description of this playbook.

        Returns:
            A string describing what this playbook does
        """
        pass

    def __repr__(self) -> str:
        """Return a string representation of the playbook."""
        return (
            f"{self.__class__.__name__}(name='{self.name}', agent='{self.agent_name}')"
        )

    def __str__(self) -> str:
        """Return a human-readable string representation."""
        return self.name

    def trigger_instructions(
        self, namespace: Optional[str] = None, skip_bgn: bool = True
    ) -> List[str]:
        """Get the trigger instructions for the playbook.

        Returns:
            A list of trigger instruction strings, or an empty list if no triggers.
        """

        instructions = []
        if self.triggers:
            for trigger in self.triggers.triggers:
                if skip_bgn and trigger.is_begin:
                    continue
                instructions.append(trigger.trigger_instruction(namespace))
        return instructions

    def create_namespace_function(self, agent) -> "Callable":
        """Create a call-through function for cross-playbook calls.

        Args:
            agent: The agent that owns this playbook

        Returns:
            Callable: A function that routes calls to agent.execute_playbook()
        """

        def call_through_agent(*args, _agent=agent, **kwargs):
            return _agent.execute_playbook(self.name, args, kwargs)

        return call_through_agent



from dataclasses import dataclass, field
from datetime import datetime
from typing import Any, List


@dataclass(frozen=True)
class Event:
    """Base class for all events."""

    session_id: str
    agent_id: str = ""
    timestamp: datetime = field(default_factory=datetime.now)


@dataclass(frozen=True)
class CallStackPushEvent(Event):
    """Call stack frame pushed."""

    frame: str = ""
    stack: List[str] = field(default_factory=list)


@dataclass(frozen=True)
class CallStackPopEvent(Event):
    """Call stack frame popped."""

    frame: str = ""
    stack: List[str] = field(default_factory=list)


@dataclass(frozen=True)
class InstructionPointerEvent(Event):
    """Instruction pointer moved."""

    pointer: str = ""
    stack: List[str] = field(default_factory=list)


@dataclass(frozen=True)
class CompiledProgramEvent(Event):
    """Program compiled successfully."""

    compiled_file_path: str = ""
    content: str = ""
    original_file_paths: List[str] = field(default_factory=list)


@dataclass(frozen=True)
class ProgramTerminatedEvent(Event):
    """Program terminated."""

    reason: str = ""
    exit_code: int = 0


@dataclass(frozen=True)
class AgentStartedEvent(Event):
    """Agent started."""

    agent_name: str = ""
    agent_type: str = ""


@dataclass(frozen=True)
class AgentStoppedEvent(Event):
    """Agent stopped."""

    agent_name: str = ""
    reason: str = ""


@dataclass(frozen=True)
class AgentPausedEvent(Event):
    """Agent paused execution."""

    reason: str = ""
    source_line_number: int = 0
    step: str = ""


@dataclass(frozen=True)
class AgentResumedEvent(Event):
    """Agent resumed execution."""

    pass


@dataclass(frozen=True)
class AgentStepEvent(Event):
    """Agent performed step operation."""

    step_mode: Any = None


@dataclass(frozen=True)
class BreakpointHitEvent(Event):
    """Breakpoint was hit."""

    file_path: str = ""
    line_number: int = 0
    source_line_number: int = 0


@dataclass(frozen=True)
class StepCompleteEvent(Event):
    """Step operation completed."""

    source_line_number: int = 0


@dataclass(frozen=True)
class VariableUpdateEvent(Event):
    """Agent variables updated."""

    variable_name: str = ""
    variable_value: Any = None


@dataclass(frozen=True)
class ExecutionPausedEvent(Event):
    """Execution paused."""

    reason: str = ""
    source_line_number: int = 0
    step: str = ""


@dataclass(frozen=True)
class LineExecutedEvent(Event):
    """Line of code executed."""

    step: str = ""
    source_line_number: int = 0
    text: str = ""
    file_path: str = ""
    line_number: int = 0


@dataclass(frozen=True)
class PlaybookStartEvent(Event):
    """Playbook started."""

    playbook: str = ""


@dataclass(frozen=True)
class PlaybookEndEvent(Event):
    """Playbook ended."""

    playbook: str = ""
    return_value: Any = None
    call_stack_depth: int = 0



from .ai_agent import AIAgent
from .base_agent import BaseAgent
from .human_agent import HumanAgent
from .local_ai_agent import LocalAIAgent
from .mcp_agent import MCPAgent
from .remote_ai_agent import RemoteAIAgent

__all__ = [
    "BaseAgent",
    "HumanAgent",
    "AIAgent",
    "LocalAIAgent",
    "RemoteAIAgent",
    "MCPAgent",
]



#!/usr/bin/env python
"""
Command line interface for the playbooks framework.

Provides commands for running and compiling playbooks.
"""
import argparse
import asyncio
import importlib
import json
import os
import sys
import warnings
from typing import Any, List

import frontmatter
import openai
from rich.console import Console
from rich.panel import Panel
from rich.table import Table

from .compiler import Compiler, FileCompilationSpec
from .exceptions import ProgramLoadError
from .loader import Loader
from .logging_setup import configure_logging
from .utils.llm_config import LLMConfig
from .utils.version import get_playbooks_version

# Suppress deprecation warnings from external libraries
warnings.filterwarnings("ignore", category=DeprecationWarning, module="httpx")
warnings.filterwarnings("ignore", category=DeprecationWarning, module="litellm")
warnings.filterwarnings("ignore", message="Accessing the 'model_fields' attribute")
warnings.filterwarnings("ignore", message="The `dict` method is deprecated")
warnings.filterwarnings(
    "ignore", message="Use 'content=<...>' to upload raw bytes/text content"
)

console = Console()


def compile(program_paths: List[str], output_file: str = None) -> None:
    """
    Compile a playbook file.

    Args:
        program_paths: List of Playbooks program files to compile
        output_file: Optional path to save compiled output. If None, prints to stdout.
    """
    if isinstance(program_paths, str):
        program_paths = [program_paths]

    # Load files individually
    program_file_tuples = Loader.read_program_files(program_paths)
    program_files = [
        FileCompilationSpec(file_path=fp, content=content, is_compiled=is_comp)
        for fp, content, is_comp in program_file_tuples
    ]

    # Let compiler handle all compilation logic
    llm_config = LLMConfig()
    compiler = Compiler(llm_config)
    compiled_results = compiler.process_files(program_files)

    try:
        for result in compiled_results:
            # Add frontmatter back to content if present
            content = result.content
            if result.frontmatter_dict:
                fm_post = frontmatter.Post(content, **result.frontmatter_dict)
                content = frontmatter.dumps(fm_post)

            if output_file:
                # Save to specified output file
                if len(compiled_results) > 1:
                    raise Exception(
                        "Do not specify output file name when compiling multiple files"
                    )

                with open(output_file, "w") as f:
                    f.write(content)
                console.print(
                    f"[green]Compiled Playbooks program saved to:[/green] {output_file}"
                )
            else:
                print(content)

    except Exception as e:
        console.print(f"[bold red]Error compiling Playbooks program:[/bold red] {e}")
        sys.exit(1)


async def run_application(
    application_module: str,
    program_paths: List[str],
    verbose: bool = False,
    stream: bool = True,
    enable_debug: bool = False,
    debug_host: str = "127.0.0.1",
    debug_port: int = 7529,
    wait_for_client: bool = False,
    stop_on_entry: bool = False,
) -> None:
    """
    Run a playbook using the specified application.

    Args:
        application_module: Module path like 'playbooks.applications.agent_chat'
        program_paths: List of playbook files to run
        verbose: Whether to print the session log
        enable_debug: Whether to start the debug server
        debug_host: Host address for the debug server
        debug_port: Port for the debug server
        wait_for_client: Whether to wait for a client to connect before starting
        stop_on_entry: Whether to stop at the beginning of playbook execution
    """
    # Import the application module
    try:
        module = importlib.import_module(application_module)
    except ModuleNotFoundError as e:
        console.print(f"[bold red]Error importing application:[/bold red] {e}")
        sys.exit(1)

    if isinstance(program_paths, str):
        program_paths = [program_paths]

    try:
        await module.main(
            program_paths=program_paths,
            verbose=verbose,
            stream=stream,
            enable_debug=enable_debug,
            debug_host=debug_host,
            debug_port=debug_port,
            wait_for_client=wait_for_client,
            stop_on_entry=stop_on_entry,
        )

    except ImportError as e:
        console.print(f"[bold red]Error importing application:[/bold red] {e}")
        console.print(
            f"[yellow]Make sure the module path is correct: {application_module}[/yellow]"
        )
        sys.exit(1)
    except Exception as e:
        import traceback

        console.print(f"[bold red]Error running application:[/bold red] {e}")
        console.print("[bold red]Traceback:[/bold red]")
        traceback.print_exc()
        sys.exit(1)


# -------------------------
# config show implementation
# -------------------------


def _mask_secrets_inplace(obj: Any) -> Any:
    """
    Recursively mask values whose keys look secret-ish.
    Heuristic: key contains one of KEY, TOKEN, SECRET, PASSWORD (case-insensitive).
    """
    if isinstance(obj, dict):
        masked = {}
        for k, v in obj.items():
            if isinstance(k, str) and any(
                s in k.lower() for s in ("key", "token", "secret", "password")
            ):
                masked[k] = "********"
            else:
                masked[k] = _mask_secrets_inplace(v)
        return masked
    if isinstance(obj, list):
        return [_mask_secrets_inplace(v) for v in obj]
    return obj


def _print_config_pretty(effective: dict, files_used: list[str], mask: bool) -> None:
    """Pretty render config and file list using rich."""
    if mask:
        effective = _mask_secrets_inplace(effective)

    table = Table(title="Effective Playbooks Configuration", show_lines=False)
    table.add_column("Key", style="cyan", no_wrap=True)
    table.add_column("Value", style="white")

    def walk(prefix: str, node: Any):
        if isinstance(node, dict):
            for k in sorted(node.keys()):
                walk(f"{prefix}.{k}" if prefix else k, node[k])
        else:
            # Render scalars/arrays as JSON for readability
            table.add_row(prefix, json.dumps(node, ensure_ascii=False))

    walk("", effective)
    console.print(table)

    if files_used:
        files_panel = Panel.fit(
            "\n".join(files_used),
            title="Files used (lowest → highest among files)",
            border_style="magenta",
        )
        console.print(files_panel)


def _cmd_config_show(args) -> None:
    # Lazy import so CLI still works if config module is missing in other contexts
    try:
        from .config import load_config
    except Exception as e:
        console.print(
            "[bold red]Unable to load configuration module (.config).[/bold red]\n"
            "Make sure `config.py` (with load_config) is available in the package."
        )
        console.print(f"[red]Detail:[/red] {e}")
        sys.exit(1)

    profile = args.profile
    explicit_path = args.config_path
    overrides = {}  # reserved for future: map additional CLI flags to schema here

    try:
        config, files = load_config(
            profile=profile,
            explicit_path=explicit_path,
            overrides=overrides,
        )
    except Exception as e:
        console.print(f"[bold red]Config error:[/bold red] {e}")
        sys.exit(1)

    effective = config.model_dump()

    if args.json:
        if args.mask_secrets:
            effective = _mask_secrets_inplace(effective)
        # Include files in JSON if requested
        payload = {"config": effective, "files_used": [str(p) for p in files]}
        console.print_json(json.dumps(payload, ensure_ascii=False, indent=2))
        return

    # Pretty (default)
    _print_config_pretty(effective, [str(p) for p in files], args.mask_secrets)


def main():
    """Main CLI entry point."""
    # Configure logging early
    configure_logging()

    parser = argparse.ArgumentParser(
        description="Playbooks CLI - Compile and run Playbooks programs",
        prog="playbooks",
    )

    # Add version argument
    parser.add_argument(
        "--version", action="version", version=f"playbooks {get_playbooks_version()}"
    )

    subparsers = parser.add_subparsers(dest="command", help="Available commands")

    # Run command
    run_parser = subparsers.add_parser("run", help="Run a Playbooks program")
    run_parser.add_argument(
        "program_paths",
        help="One or more paths to the Playbooks program files to run",
        nargs="+",
    )
    run_parser.add_argument(
        "--application",
        default="playbooks.applications.agent_chat",
        help="Application module to use (default: playbooks.applications.agent_chat)",
    )
    run_parser.add_argument(
        "-v", "--verbose", action="store_true", help="Print the session log"
    )
    run_parser.add_argument(
        "--stream",
        type=lambda x: x.lower() in ["true", "1", "yes"],
        default=True,
        help="Enable/disable streaming output (default: True). Use --stream=False for buffered output",
    )
    run_parser.add_argument(
        "--debug", action="store_true", help="Start the debug server"
    )
    run_parser.add_argument(
        "--debug-host",
        default="127.0.0.1",
        help="Debug server host (default: 127.0.0.1)",
    )
    run_parser.add_argument(
        "--debug-port", type=int, default=7529, help="Debug server port (default: 7529)"
    )
    run_parser.add_argument(
        "--wait-for-client",
        action="store_true",
        help="Wait for a debug client to connect before starting execution",
    )
    run_parser.add_argument(
        "--skip-compilation",
        action="store_true",
        help="Skip compilation step (skipped automatically for .pbasm files)",
    )
    run_parser.add_argument(
        "--stop-on-entry",
        action="store_true",
        help="Stop at the beginning of playbook execution",
    )

    # Compile command
    compile_parser = subparsers.add_parser("compile", help="Compile a playbook")
    compile_parser.add_argument(
        "program_paths",
        help="One or more paths to the playbook files to compile",
        nargs="+",
    )
    compile_parser.add_argument(
        "--output", help="Output file path (if not specified, prints to stdout)"
    )

    # Webserver command
    webserver_parser = subparsers.add_parser(
        "webserver", help="Start the Playbooks web server"
    )
    webserver_parser.add_argument(
        "--host", default="localhost", help="Host address (default: localhost)"
    )
    webserver_parser.add_argument(
        "--http-port", type=int, default=8000, help="HTTP port (default: 8000)"
    )
    webserver_parser.add_argument(
        "--ws-port", type=int, default=8001, help="WebSocket port (default: 8001)"
    )

    # Playground command
    playground_parser = subparsers.add_parser(
        "playground", help="Start the Playbooks playground (webserver + browser)"
    )
    playground_parser.add_argument(
        "--host", default="localhost", help="Host address (default: localhost)"
    )
    playground_parser.add_argument(
        "--http-port", type=int, default=8000, help="HTTP port (default: 8000)"
    )
    playground_parser.add_argument(
        "--ws-port", type=int, default=8001, help="WebSocket port (default: 8001)"
    )

    # -------------------------
    # Config command group
    # -------------------------
    config_parser = subparsers.add_parser(
        "config", help="Inspect and manage Playbooks configuration"
    )
    cfg_sub = config_parser.add_subparsers(dest="config_cmd", required=True)

    cfg_show = cfg_sub.add_parser(
        "show", help="Show effective configuration and sources"
    )
    cfg_show.add_argument(
        "--profile",
        default=os.getenv("PLAYBOOKS_PROFILE"),
        help="Profile name to apply (e.g., prod). Defaults to $PLAYBOOKS_PROFILE.",
    )
    cfg_show.add_argument(
        "--config",
        dest="config_path",
        help="Explicit config file to load in addition to project/user files.",
    )
    cfg_show.add_argument(
        "--mask-secrets",
        action="store_true",
        help="Mask likely secrets (keys containing key/token/secret/password).",
    )
    cfg_show.add_argument(
        "--json",
        action="store_true",
        help="Output JSON instead of a table (includes files_used).",
    )

    args = parser.parse_args()

    if not args.command:
        parser.print_help()
        sys.exit(1)

    if args.command == "run":
        # Run the application
        try:
            asyncio.run(
                run_application(
                    args.application,
                    args.program_paths,
                    verbose=args.verbose,
                    stream=args.stream,
                    enable_debug=args.debug,
                    debug_host=args.debug_host,
                    debug_port=args.debug_port,
                    wait_for_client=args.wait_for_client,
                    stop_on_entry=args.stop_on_entry,
                )
            )
        except KeyboardInterrupt:
            console.print("\n[yellow]Interrupted by user[/yellow]")
        except ProgramLoadError as e:
            console.print(f"[bold red]Error loading program:[/bold red] {e}")
            sys.exit(1)

    elif args.command == "compile":
        try:
            compile(
                args.program_paths,
                args.output,
            )
        except KeyboardInterrupt:
            console.print("\n[yellow]Interrupted by user[/yellow]")
        except openai.OpenAIError:
            import traceback

            traceback.print_exc()
            console.print(
                "[bold red]Error: Authentication failed. Please make sure you have a valid ANTHROPIC_API_KEY environment variable set.[/bold red]"
            )
            sys.exit(1)
        except ProgramLoadError as e:
            console.print(f"[bold red]Error loading program:[/bold red] {e}")
            sys.exit(1)
        except Exception as e:
            console.print(f"[bold red]Error compiling playbooks:[/bold red] {e}")
            sys.exit(1)

    elif args.command == "webserver":
        try:
            # Import here to avoid unnecessary imports if not using webserver
            from .applications.web_server import PlaybooksWebServer

            async def start_webserver():
                server = PlaybooksWebServer(args.host, args.http_port, args.ws_port)
                await server.start()

            asyncio.run(start_webserver())
        except KeyboardInterrupt:
            console.print("\n[yellow]Web server stopped[/yellow]")
        except Exception as e:
            console.print(f"[bold red]Error starting web server:[/bold red] {e}")
            sys.exit(1)

    elif args.command == "playground":
        try:
            import threading
            import time
            import webbrowser
            from pathlib import Path

            # Import here to avoid unnecessary imports if not using webserver
            from .applications.web_server import PlaybooksWebServer

            # Get the path to the playground HTML file
            playground_path = (
                Path(__file__).parent / "applications" / "playbooks_playground.html"
            )
            playground_url = f"file://{playground_path.absolute()}"

            console.print("[green]Starting Playbooks Playground...[/green]")
            console.print(
                f"[cyan]Server will start on:[/cyan] http://{args.host}:{args.http_port}"
            )
            console.print(
                f"[cyan]WebSocket will start on:[/cyan] ws://{args.host}:{args.ws_port}"
            )

            # Function to open browser after a short delay
            def open_browser():
                time.sleep(2)  # Give server time to start
                console.print(
                    f"[green]Opening playground in browser:[/green] {playground_url}"
                )
                webbrowser.open(playground_url)

            # Start browser opener in background thread
            browser_thread = threading.Thread(target=open_browser, daemon=True)
            browser_thread.start()

            async def start_playground():
                server = PlaybooksWebServer(args.host, args.http_port, args.ws_port)
                console.print(f"[yellow]Playground URL:[/yellow] {playground_url}")
                console.print("[dim]Press Ctrl+C to stop[/dim]")
                await server.start()

            asyncio.run(start_playground())

        except KeyboardInterrupt:
            console.print("\n[yellow]Playground stopped[/yellow]")
        except Exception as e:
            console.print(f"[bold red]Error starting playground:[/bold red] {e}")
            sys.exit(1)

    elif args.command == "config":
        if args.config_cmd == "show":
            _cmd_config_show(args)
        else:
            console.print("[red]Unknown config subcommand[/red]")
            sys.exit(2)


if __name__ == "__main__":
    main()



"""React execution that loops until exit conditions are met."""

from typing import TYPE_CHECKING, Any

from ..compiler import Compiler
from ..playbook_step import PlaybookStep, PlaybookStepCollection
from ..utils.llm_config import LLMConfig
from ..utils.markdown_to_ast import markdown_to_ast
from .playbook import PlaybookLLMExecution

if TYPE_CHECKING:
    pass


class ReActLLMExecution(PlaybookLLMExecution):
    """React execution that loops until exit conditions are met.

    This mode enables the LLM to repeatedly:
    - Think about the current state
    - Decide on actions (tool calls, user interaction)
    - Execute those actions
    - Check exit conditions
    - Continue or complete
    """

    async def execute(self, *args, **kwargs) -> Any:
        """Execute with react looping behavior."""
        # Add react loop steps if not present
        if not self._has_steps():
            self._add_react_steps()

        # Delegate to parent's playbook execution
        return await super().execute(*args, **kwargs)

    def _has_steps(self) -> bool:
        """Check if the playbook has steps defined."""
        return (
            hasattr(self.playbook, "step_collection")
            and self.playbook.step_collection
            and len(self.playbook.step_collection.ordered_line_numbers) > 0
        )

    def _add_react_steps(self) -> None:
        """Add react loop steps to the playbook."""
        # Create a ReactAgent playbook for compilation
        react_playbook = """# ReactAgent

## ReactSteps
### Steps
- Think deeply about the $task to understand requirements
- Write down $exit_conditions for the task
- While $exit_conditions are not met:
  - Analyze current state and progress
  - Decide what action to take next
  - Execute the action (tool call, user interaction, computation)
  - Evaluate results against exit conditions
- Return final results
"""

        # Compile the playbook to get proper step structure
        compiled_steps = self._compile_and_extract_steps(react_playbook)

        # Set the compiled step collection on the playbook
        self.playbook.step_collection = compiled_steps

    def _compile_and_extract_steps(
        self, playbook_content: str
    ) -> PlaybookStepCollection:
        """Compile the playbook content and extract step collection.

        Args:
            playbook_content: The markdown playbook content to compile

        Returns:
            PlaybookStepCollection: The compiled step collection
        """
        try:
            # Create compiler with LLM config
            llm_config = LLMConfig()
            compiler = Compiler(llm_config)

            # Compile the playbook content
            _, compiled_content, _ = compiler.compile(content=playbook_content)

            # Parse the compiled content to extract steps
            ast = markdown_to_ast(compiled_content)

            # Find the ReactAgent H1 section
            react_agent_h1 = None
            for child in ast.get("children", []):
                if child.get("type") == "h1" and "ReactAgent" in child.get("text", ""):
                    react_agent_h1 = child
                    break

            if not react_agent_h1:
                # Fallback to original behavior if compilation failed
                return self._create_fallback_steps()

            # Find the ReactSteps H2 section within the ReactAgent
            react_steps_h2 = None
            for child in react_agent_h1.get("children", []):
                if child.get("type") == "h2" and "ReactSteps" in child.get("text", ""):
                    react_steps_h2 = child
                    break

            if not react_steps_h2:
                # Fallback to original behavior if H2 not found
                return self._create_fallback_steps()

            # Extract steps from the compiled H2 section
            return self._extract_steps_from_h2(react_steps_h2)

        except Exception as e:
            # If any error occurs during compilation or parsing,
            # fall back to the original implementation
            raise Exception(f"Error creating react steps: {e}") from e

    def _extract_steps_from_h2(self, h2_node: dict) -> PlaybookStepCollection:
        """Extract steps from a compiled H2 node.

        Args:
            h2_node: The H2 AST node containing steps

        Returns:
            PlaybookStepCollection: The extracted step collection
        """

        def parse_node(
            node: dict, step_collection: PlaybookStepCollection
        ) -> PlaybookStep:
            step = None
            if node.get("type") == "list-item":
                text = node.get("text", "").strip()
                item_line_number = node.get("line_number")
                source_file_path = node.get("source_file_path")
                step = PlaybookStep.from_text(text)
                if step:
                    step.source_line_number = item_line_number
                    step.source_file_path = source_file_path
                    step_collection.add_step(step)

                    if node.get("children"):
                        if len(node.get("children")) > 1:
                            raise ValueError(
                                f"Expected 1 child for list-item, got {len(node.get('children'))}"
                            )

                        list_node = node.get("children")[0]
                        if list_node.get("type") != "list":
                            raise ValueError(
                                f"Expected a single list under list-item, got a {list_node.get('type')}"
                            )

                        child_steps = []
                        for child in list_node.get("children", []):
                            child_step = parse_node(child, step_collection)
                            if child_step:
                                child_steps.append(child_step)

                        step.children = child_steps
            else:
                raise ValueError(f"Expected a list-item, got a {node.get('type')}")
            return step

        # Find the Steps H3 section
        for child in h2_node.get("children", []):
            if (
                child.get("type") == "h3"
                and child.get("text", "").strip().lower() == "steps"
            ):
                step_collection = PlaybookStepCollection()
                if child["children"] and "children" in child["children"][0]:
                    for list_item in child["children"][0]["children"]:
                        parse_node(list_item, step_collection)
                return step_collection

        # No steps found, return empty collection
        return PlaybookStepCollection()



import json
import os
from typing import TYPE_CHECKING, Dict, List, Optional

from playbooks.debug_logger import debug
from playbooks.llm_context_compactor import LLMContextCompactor
from playbooks.llm_messages import (
    AgentInfoLLMMessage,
    AssistantResponseLLMMessage,
    OtherAgentInfoLLMMessage,
    TriggerInstructionsLLMMessage,
    UserInputLLMMessage,
)
from playbooks.playbook import Playbook
from playbooks.utils.llm_helper import get_messages_for_prompt
from playbooks.utils.token_counter import get_messages_token_count

if TYPE_CHECKING:
    from playbooks.execution_state import ExecutionState


class InterpreterPrompt:
    """Generates the prompt for the interpreter LLM based on the current state."""

    def __init__(
        self,
        execution_state: "ExecutionState",
        playbooks: Dict[str, Playbook],
        current_playbook: Optional[Playbook],
        instruction: str,
        agent_instructions: str,
        artifacts_to_load: List[str],
        trigger_instructions: List[str],
        agent_information: str,
        other_agent_klasses_information: List[str],
    ):
        """
        Initializes the InterpreterPrompt.

        Args:
            execution_state: The current execution state.
            playbooks: A dictionary of available playbooks.
            current_playbook: The currently executing playbook, if any.
            instruction: The user's latest instruction.
            agent_instructions: General instructions for the agent.
            artifacts_to_load: List of artifact names to load.
        """
        self.execution_state = execution_state
        self.playbooks = playbooks
        self.current_playbook = current_playbook
        self.instruction = instruction
        self.agent_instructions = agent_instructions
        self.artifacts_to_load = artifacts_to_load
        self.trigger_instructions = trigger_instructions
        self.agent_information = agent_information
        self.other_agent_klasses_information = other_agent_klasses_information
        self.compactor = LLMContextCompactor()

    def _get_trigger_instructions_message(self) -> str:
        if len(self.trigger_instructions) > 0:
            trigger_instructions = (
                ["*Available playbook triggers*", "```md"]
                + self.trigger_instructions
                + ["```"]
            )

            return TriggerInstructionsLLMMessage(
                "\n".join(trigger_instructions)
            ).to_full_message()
        return None

    def _get_other_agent_klasses_information_message(self) -> str:
        if len(self.other_agent_klasses_information) > 0:
            other_agent_klasses_information = [
                "*Other agents*",
                "```md",
                "\n\n".join(self.other_agent_klasses_information),
                "```",
            ]
            return OtherAgentInfoLLMMessage(
                "\n".join(other_agent_klasses_information)
            ).to_full_message()
        return None

    def _get_compact_agent_information_message(self) -> str:
        parts = []
        parts.append("*My agent*")
        parts.append("```md")
        parts.append(self.agent_information)
        parts.append("```")
        return AgentInfoLLMMessage("\n".join(parts)).to_full_message()

    @property
    def prompt(self) -> str:
        """Constructs the full prompt string for the LLM.

        Returns:
            The formatted prompt string.
        """
        # trigger_instructions_str = self._get_trigger_instructions_str()

        # current_playbook_markdown = (
        #     self.playbooks[self.current_playbook.klass].markdown
        #     if self.current_playbook
        #     else "No playbook is currently running."
        # )

        try:
            with open(
                os.path.join(
                    os.path.dirname(__file__), "./prompts/interpreter_run.txt"
                ),
                "r",
            ) as f:
                prompt = f.read()
        except FileNotFoundError:
            debug("Error: Prompt template file not found")
            return "Error: Prompt template missing."

        initial_state = json.dumps(self.execution_state.to_dict(), indent=2)

        # session_log_str = str(self.execution_state.session_log)

        # prompt = prompt_template.replace("{{TRIGGERS}}", trigger_instructions_str)
        # prompt = prompt.replace(
        #     "{{CURRENT_PLAYBOOK_MARKDOWN}}", current_playbook_markdown
        # )
        # prompt = prompt.replace("{{SESSION_LOG}}", session_log_str)
        prompt = prompt.replace("{{INITIAL_STATE}}", initial_state)
        prompt = prompt.replace("{{INSTRUCTION}}", self.instruction)
        if self.agent_instructions:
            prompt = prompt.replace("{{AGENT_INSTRUCTIONS}}", self.agent_instructions)
        else:
            prompt = prompt.replace("{{AGENT_INSTRUCTIONS}}", "")
        return prompt

    @property
    def messages(self) -> List[Dict[str, str]]:
        """Formats the prompt into the message structure expected by the LLM helper."""
        prompt_messages = get_messages_for_prompt(self.prompt)

        messages = []
        messages.append(prompt_messages[0])

        other_agent_klasses_information_message = (
            self._get_other_agent_klasses_information_message()
        )
        if other_agent_klasses_information_message:
            messages.append(other_agent_klasses_information_message)

        messages.append(self._get_compact_agent_information_message())

        trigger_instructions_message = self._get_trigger_instructions_message()
        if trigger_instructions_message:
            messages.append(trigger_instructions_message)

        # Convert the prompt message dict back to a proper message object
        if len(prompt_messages) > 1:
            user_instruction_msg = UserInputLLMMessage(prompt_messages[1]["content"])
            self.execution_state.call_stack.add_llm_message(user_instruction_msg)

        # Original call stack messages (as LLMMessage objects)
        call_stack_llm_messages = []
        for frame in self.execution_state.call_stack.frames:
            call_stack_llm_messages.extend(frame.llm_messages)

        # Apply compaction - returns Dict[str, str] format ready for LLM API
        compacted_dict_messages = self.compactor.compact_messages(
            call_stack_llm_messages
        )

        # Log compaction stats using token counts
        original_dict_messages = [
            msg.to_full_message() for msg in call_stack_llm_messages
        ]
        original_tokens = get_messages_token_count(messages + original_dict_messages)
        compacted_tokens = get_messages_token_count(messages + compacted_dict_messages)
        compression_ratio = (
            compacted_tokens / original_tokens if original_tokens > 0 else 1.0
        )

        debug(
            f"LLM Context Compaction: {original_tokens} -> {compacted_tokens} tokens ({compression_ratio:.2%})"
        )

        messages.extend(compacted_dict_messages)
        # messages.extend(self._get_artifact_messages())
        # messages.append(prompt_messages[1])

        return messages

    def _get_artifact_messages(self) -> List[Dict[str, str]]:
        """Generates messages for the artifacts to load."""
        artifact_messages = []
        for artifact in self.artifacts_to_load:
            artifact = self.execution_state.artifacts[artifact]
            artifact_message = f"Artifact[{artifact.name}]\n\nSummary: {artifact.summary}\n\nContent: {artifact.content}"
            artifact_messages.append(
                AssistantResponseLLMMessage(artifact_message).to_full_message()
            )
        return artifact_messages



import asyncio
import logging
from collections import defaultdict
from typing import Awaitable, Callable, Dict, List, Type, TypeVar, Union
from weakref import WeakSet

from .events import Event

logger = logging.getLogger(__name__)

T = TypeVar("T", bound=Event)


class EventBus:
    """Event bus for typed events with sync and async support."""

    def __init__(self, session_id: str):
        self.session_id = session_id
        self._handlers: Dict[Type[Event], List[Callable]] = defaultdict(list)
        self._global_handlers: List[Callable] = []
        self._active_tasks: WeakSet[asyncio.Task] = WeakSet()
        self._closing = False

    def subscribe(
        self,
        event_type: Union[Type[T], str],
        callback: Callable[[T], Union[None, Awaitable[None]]],
    ) -> None:
        """Subscribe to events of a specific type."""
        if self._closing:
            raise RuntimeError("Cannot subscribe to closing event bus")

        if isinstance(event_type, str) and event_type == "*":
            self._global_handlers.append(callback)
        else:
            self._handlers[event_type].append(callback)

    def unsubscribe(
        self,
        event_type: Union[Type[T], str],
        callback: Callable[[T], Union[None, Awaitable[None]]],
    ) -> None:
        """Remove a previously registered callback."""
        try:
            if isinstance(event_type, str) and event_type == "*":
                self._global_handlers.remove(callback)
            else:
                self._handlers[event_type].remove(callback)
                if not self._handlers[event_type]:
                    del self._handlers[event_type]
        except ValueError:
            pass

    def publish(self, event: Event) -> None:
        """Publish an event synchronously."""
        # Events are frozen, so session_id should be set during construction

        # Get handlers
        callbacks = list(self._handlers.get(type(event), []))
        callbacks.extend(self._global_handlers)

        # Execute callbacks
        for callback in callbacks:
            try:
                result = callback(event)
                if asyncio.iscoroutine(result):
                    # Schedule async callback
                    try:
                        loop = asyncio.get_running_loop()
                        task = loop.create_task(result)
                        self._active_tasks.add(task)
                    except RuntimeError:
                        # No event loop running
                        asyncio.run(result)
            except Exception as e:
                logger.error(
                    f"Error in subscriber for {type(event).__name__}: {e}",
                    exc_info=True,
                )

    async def publish_async(self, event: Event) -> None:
        """Publish an event asynchronously."""
        if self._closing:
            raise RuntimeError("Cannot publish to closing event bus")

        # Events are frozen, so session_id should be set during construction

        event_type = type(event)

        # Get handlers
        handlers = list(self._handlers.get(event_type, []))
        handlers.extend(self._global_handlers)

        if not handlers:
            return

        # Execute handlers concurrently
        tasks = []
        for handler in handlers:
            task = asyncio.create_task(self._safe_handler(handler, event))
            tasks.append(task)
            self._active_tasks.add(task)

        # Wait for completion with error isolation
        results = await asyncio.gather(*tasks, return_exceptions=True)

        for i, result in enumerate(results):
            if isinstance(result, Exception):
                logger.error(
                    f"Handler {i} failed for {event_type.__name__}: {result}",
                    exc_info=True,
                )

    async def _safe_handler(self, handler: Callable, event: Event) -> None:
        """Execute handler with error isolation."""
        try:
            result = handler(event)
            if asyncio.iscoroutine(result):
                await result
        except asyncio.CancelledError:
            raise
        except Exception as e:
            logger.error(f"Exception in handler {handler.__name__}: {e}", exc_info=True)
            raise

    def clear_subscribers(self, event_type: Type[Event] = None) -> None:
        """Clear all subscribers or subscribers of a specific event type."""
        if event_type:
            self._handlers.pop(event_type, None)
        else:
            self._handlers.clear()
            self._global_handlers.clear()

    async def close(self) -> None:
        """Close the event bus gracefully."""
        self._closing = True

        # Cancel active tasks
        active_tasks = list(self._active_tasks)
        for task in active_tasks:
            if not task.done():
                task.cancel()

        # Wait for cleanup
        if active_tasks:
            try:
                await asyncio.wait_for(
                    asyncio.gather(*active_tasks, return_exceptions=True), timeout=5.0
                )
            except asyncio.TimeoutError:
                logger.warning("Some handlers did not complete during shutdown")

        self.clear_subscribers()

    async def __aenter__(self):
        """Context manager entry."""
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        """Context manager exit with cleanup."""
        await self.close()

    @property
    def subscriber_count(self) -> Dict[Union[Type[Event], str], int]:
        """Get count of subscribers per event type (including wildcard)."""
        counts: Dict[Union[Type[Event], str], int] = {
            event_type: len(callbacks)
            for event_type, callbacks in self._handlers.items()
        }
        if self._global_handlers:
            counts["*"] = len(self._global_handlers)
        return counts

    @property
    def is_closing(self) -> bool:
        """Check if event bus is closing."""
        return self._closing



"""MCP transport implementation using FastMCP client."""

import logging
from typing import Any, Dict, List, Optional

from fastmcp import Client
from fastmcp.client.transports import PythonStdioTransport, SSETransport

from .protocol import TransportProtocol

logger = logging.getLogger(__name__)


class MCPTransport(TransportProtocol):
    """Transport implementation for MCP (Model Context Protocol) using FastMCP.

    This transport can connect to MCP servers using various underlying transports
    like SSE (Server-Sent Events), stdio, or WebSocket.
    """

    def __init__(self, config: Dict[str, Any]):
        """Initialize MCP transport.

        Args:
            config: Configuration dictionary containing:
                - url: The MCP server URL or command
                - transport: Transport type ('sse', 'stdio', 'websocket', etc.)
                - auth: Optional authentication configuration
                - timeout: Optional timeout in seconds
        """
        super().__init__(config)
        self.url = config.get("url")
        self.transport_type = config.get("transport", "sse")  # Default to SSE
        self.auth = config.get("auth", {})
        self.timeout = config.get("timeout", 30.0)

        self.client: Optional[Client] = None
        self._tools_cache: Optional[List[Dict[str, Any]]] = None

        if not self.url:
            raise ValueError("MCP transport requires 'url' in configuration")

    async def connect(self) -> None:
        """Establish connection to the MCP server."""
        if self._connected:
            return

        try:
            logger.debug(
                f"Connecting to MCP server at {self.url} using {self.transport_type}"
            )

            # Create the appropriate transport based on configuration
            if self.transport_type.lower() == "sse":
                transport = SSETransport(self.url)
                self.client = Client(transport)
            elif self.transport_type.lower() == "stdio":
                # For stdio, the URL should be a command/script path
                transport = PythonStdioTransport(self.url)
                self.client = Client(transport)
            else:
                # Let FastMCP auto-detect the transport
                self.client = Client(self.url)

            # Connect to the server
            await self.client.__aenter__()
            self._connected = True

            logger.info(f"Successfully connected to MCP server at {self.url}")

        except Exception as e:
            logger.error(f"Failed to connect to MCP server at {self.url}: {str(e)}")
            raise ConnectionError(f"MCP connection failed: {str(e)}") from e

    async def disconnect(self) -> None:
        """Close connection to the MCP server."""
        if not self._connected or not self.client:
            return

        try:
            await self.client.__aexit__(None, None, None)
            logger.debug(f"Disconnected from MCP server at {self.url}")
        except Exception as e:
            logger.warning(f"Error during MCP disconnect: {str(e)}")
        finally:
            self.client = None
            self._connected = False
            self._tools_cache = None

    async def call(self, method: str, params: Optional[Dict[str, Any]] = None) -> Any:
        """Make a remote call using MCP protocol.

        Args:
            method: The method to call. For MCP, this can be:
                - 'list_tools': List available tools
                - 'call_tool': Call a specific tool (requires tool_name in params)
                - 'list_resources': List available resources
                - 'read_resource': Read a specific resource (requires uri in params)
                - 'get_prompt': Get a prompt (requires name in params)
            params: Parameters for the method call

        Returns:
            The result of the MCP call

        Raises:
            ConnectionError: If not connected
            ValueError: If method or params are invalid
            Exception: If the MCP call fails
        """
        if not self._connected or not self.client:
            raise ConnectionError("MCP transport not connected")

        if not method:
            raise ValueError("Method name is required")

        params = params or {}

        try:
            logger.debug(f"MCP call: {method} with params: {params}")

            if method == "list_tools":
                result = await self.client.list_tools()
                # Cache tools for later use
                self._tools_cache = result
                return result

            elif method == "call_tool":
                tool_name = params.get("tool_name")
                tool_args = params.get("arguments", {})
                if not tool_name:
                    raise ValueError("call_tool requires 'tool_name' parameter")
                result = await self.client.call_tool(tool_name, tool_args)
                return result

            elif method == "list_resources":
                result = await self.client.list_resources()
                return result

            elif method == "read_resource":
                uri = params.get("uri")
                if not uri:
                    raise ValueError("read_resource requires 'uri' parameter")
                result = await self.client.read_resource(uri)
                return result

            elif method == "get_prompt":
                name = params.get("name")
                prompt_args = params.get("arguments", {})
                if not name:
                    raise ValueError("get_prompt requires 'name' parameter")
                result = await self.client.get_prompt(name, prompt_args)
                return result

            elif method == "ping":
                result = await self.client.ping()
                return result

            else:
                raise ValueError(f"Unsupported MCP method: {method}")

        except Exception as e:
            logger.error(f"MCP call failed - method: {method}, error: {str(e)}")
            raise

    async def list_tools(self) -> List[Dict[str, Any]]:
        """Convenience method to list available MCP tools."""
        return await self.call("list_tools")

    async def call_tool(self, tool_name: str, arguments: Dict[str, Any]) -> Any:
        """Convenience method to call an MCP tool."""
        return await self.call(
            "call_tool", {"tool_name": tool_name, "arguments": arguments}
        )

    async def list_resources(self) -> List[Dict[str, Any]]:
        """Convenience method to list available MCP resources."""
        return await self.call("list_resources")

    async def read_resource(self, uri: str) -> Any:
        """Convenience method to read an MCP resource."""
        return await self.call("read_resource", {"uri": uri})

    def get_cached_tools(self) -> Optional[List[Dict[str, Any]]]:
        """Get cached tools list if available."""
        return self._tools_cache



#!/usr/bin/env python
"""
CLI application for interactive agent chat using playbooks.
Provides a simple terminal interface for communicating with AI agents.
"""

import argparse
import asyncio
import functools
import os
import select
import sys
import uuid
from pathlib import Path
from typing import Callable, List

# Platform-specific imports for stdin clearing
try:
    import termios
except ImportError:
    termios = None

try:
    import msvcrt
except ImportError:
    msvcrt = None

import litellm
from rich.console import Console

from playbooks import Playbooks
from playbooks.agents.messaging_mixin import MessagingMixin
from playbooks.constants import EOM, EXECUTION_FINISHED
from playbooks.debug_logger import debug
from playbooks.events import Event
from playbooks.exceptions import ExecutionFinished
from playbooks.meetings.meeting_manager import MeetingManager
from playbooks.message import MessageType
from playbooks.program import Program
from playbooks.session_log import SessionLogItemLevel
from playbooks.user_output import user_output
from playbooks.utils.error_utils import check_playbooks_health

# Add the src directory to the Python path to import playbooks
sys.path.insert(0, str(Path(__file__).parent.parent))

# Initialize Rich console
console = Console()


def clear_stdin():
    """Clear any pending input from stdin buffer.

    This prevents pre-filled input when prompting the user.
    Uses platform-specific methods for optimal clearing.
    """
    try:
        if os.name == "nt" and msvcrt is not None:  # Windows
            # Clear Windows console input buffer
            while msvcrt.kbhit():
                msvcrt.getch()
        else:  # Unix/Linux/macOS
            # Use termios for aggressive clearing if available
            if termios is not None:
                try:
                    termios.tcflush(sys.stdin, termios.TCIFLUSH)
                    return  # Success, no need for fallback
                except (OSError, AttributeError):
                    pass  # Fall through to select-based approach

            # Fallback: use select to check and clear available input
            if hasattr(select, "select"):
                # Check if there's input available without blocking
                if select.select([sys.stdin], [], [], 0.0)[0]:
                    # Read and discard available input
                    try:
                        while select.select([sys.stdin], [], [], 0.0)[0]:
                            sys.stdin.read(1)
                    except (OSError, IOError):
                        pass

    except Exception:
        # Ignore all errors - stdin clearing is a best-effort operation
        pass


class PubSub:
    """Simple publish-subscribe mechanism for event handling."""

    def __init__(self):
        self.subscribers: List[Callable] = []

    def subscribe(self, callback: Callable):
        """Subscribe a callback function to receive messages."""
        self.subscribers.append(callback)

    def publish(self, message):
        """Publish a message to all subscribers."""
        for subscriber in self.subscribers:
            subscriber(message)


class SessionLogWrapper:
    """Wrapper around session_log that publishes updates."""

    def __init__(self, session_log, pubsub, verbose=False, agent=None, stream=True):
        self._session_log = session_log
        self._pubsub = pubsub
        self.verbose = verbose
        self.agent = agent
        self.stream_enabled = stream
        self.current_streaming_panel = None
        self.streaming_content = ""
        self.is_streaming = False

    def append(self, msg, level=SessionLogItemLevel.MEDIUM):
        """Append a message to the session log and publish it."""
        self._session_log.append(msg, level)
        # Skip traditional display entirely - streaming handles all output now
        # Traditional panels are disabled in favor of streaming output

        if self.verbose:
            self._pubsub.publish(str(msg))

    def __iter__(self):
        return iter(self._session_log)

    def __str__(self):
        return str(self._session_log)

    async def start_streaming_say(self, recipient=None):
        """Start displaying a streaming Say() message."""
        agent_name = self.agent.klass if self.agent else "Agent"
        self.streaming_content = ""
        self.is_streaming = True

        if self.stream_enabled:
            # Print agent name to show streaming is starting
            console.print(f"\n[green]{agent_name}:[/green] ", end="")

    async def stream_say_update(self, content: str):
        """Add content to the current streaming Say() message."""
        self.streaming_content += content
        if self.stream_enabled:
            # Simple streaming display - just print the new content
            print(content, end="", flush=True)

    async def complete_streaming_say(self):
        """Complete the current streaming Say() message."""
        if self.is_streaming:
            if self.stream_enabled:
                # Print a new line to finish the streaming output
                console.print()
            else:
                # Non-streaming mode: display the complete message now
                agent_name = self.agent.klass if self.agent else "Agent"
                console.print(
                    f"\n[green]{agent_name}:[/green] {self.streaming_content}"
                )

            # Reset streaming state
            self.is_streaming = False
            self.streaming_content = ""


# Store original methods for restoring later
original_wait_for_message = MessagingMixin.WaitForMessage
original_broadcast_to_meeting = None  # Will be set after MeetingManager is imported
original_route_message = None  # Will be set after Program is loaded


@functools.wraps(original_wait_for_message)
async def patched_wait_for_message(self, source_agent_id: str):
    """Patched version of WaitForMessage that shows a prompt when waiting for human input."""
    # For human input, show a prompt before calling the normal WaitForMessage
    if source_agent_id == "human":
        # Check if there are already messages waiting
        messages = self._message_buffer
        human_messages = [msg for msg in messages if msg.sender_id == "human"]

        if not human_messages:
            # No human messages waiting, show prompt
            console.print()  # Add a newline for spacing
            # Clear stdin buffer to prevent pre-filled input
            await asyncio.to_thread(clear_stdin)
            user_input = await asyncio.to_thread(
                console.input, "[bold yellow]User:[/bold yellow] "
            )
            # Send the user input and EOM
            program: Program = self.program
            for message in [user_input, EOM]:
                await program.route_message(
                    sender_id="human",
                    sender_klass="human",
                    receiver_spec=f"agent {self.id}",
                    message=message,
                )

    # Call the normal WaitForMessage which handles message delivery
    return await original_wait_for_message(self, source_agent_id)


async def patched_broadcast_to_meeting_as_owner(
    self,
    meeting_id: str,
    message: str,
    from_agent_id: str = None,
    from_agent_klass: str = None,
):
    """Patched version of broadcast_to_meeting_as_owner that displays meeting messages nicely."""
    # Display the meeting message with formatting
    if not from_agent_id or not from_agent_klass:
        from_agent_id = self.agent.id
        from_agent_klass = self.agent.klass

    # Format and display the meeting broadcast
    console.print(
        f"\n[bold blue]📢 Meeting {meeting_id}[/bold blue] - [cyan]{from_agent_klass}({from_agent_id})[/cyan]: {message}"
    )

    # Call the original method
    if original_broadcast_to_meeting:
        return await original_broadcast_to_meeting(
            self, meeting_id, message, from_agent_id, from_agent_klass
        )


async def patched_route_message(
    self,
    sender_id: str,
    sender_klass: str,
    receiver_spec: str,
    message: str,
    message_type=MessageType.DIRECT,
    meeting_id: str = None,
):
    """Patched version of route_message that displays agent-to-agent messages."""
    # Extract receiver info
    from playbooks.utils.spec_utils import SpecUtils

    debug(
        "Patched route message",
        sender_id=sender_id,
        receiver_spec=receiver_spec,
        message=message[:50],
    )

    recipient_id = SpecUtils.extract_agent_id(receiver_spec)

    recipient = self.agents_by_id.get(recipient_id)
    recipient_klass = recipient.klass if recipient else "Unknown"

    # Display agent-to-agent message with formatting
    console.print(
        f"\n[bold magenta]💬 Message[/bold magenta]: [purple]{sender_klass}({sender_id})[/purple] → [purple]{recipient_klass}({recipient_id})[/purple]: {message}"
    )

    # Call the original method
    if original_route_message:
        return await original_route_message(
            self,
            sender_id,
            sender_klass,
            receiver_spec,
            message,
            message_type,
            meeting_id,
        )


async def main(
    program_paths: str,
    verbose: bool,
    enable_debug: bool = False,
    debug_host: str = "127.0.0.1",
    debug_port: int = 7529,
    wait_for_client: bool = False,
    stop_on_entry: bool = False,
    stream: bool = True,
):
    """
    Playbooks application host for agent chat. You can execute a playbooks program within this application container.

    Example:
        ```sh
        python -m playbooks.applications.agent_chat tests/data/02-personalized-greeting.pb
        ```

    Args:
        program_paths: Path to the playbook file(s) to load
        verbose: Whether to print the session log
        enable_debug: Whether to start the debug server
        debug_host: Host address for the debug server
        debug_port: Port for the debug server
        wait_for_client: Whether to wait for a client to connect before starting
        stop_on_entry: Whether to stop at the beginning of playbook execution
        stream: Whether to stream the output

    """
    #     f"[DEBUG] agent_chat.main called with stop_on_entry={stop_on_entry}, debug={debug}"
    # )

    # Patch the WaitForMessage method before loading agents
    MessagingMixin.WaitForMessage = patched_wait_for_message

    user_output.info(f"Loading playbooks from: {program_paths}")

    session_id = str(uuid.uuid4())
    if isinstance(program_paths, str):
        program_paths = [program_paths]
    try:
        playbooks = Playbooks(program_paths, session_id=session_id)
        await playbooks.initialize()
    except litellm.exceptions.AuthenticationError as e:
        user_output.error("Authentication error", details=str(e))
        raise

    # Store original methods and apply patches after playbooks are loaded
    global original_broadcast_to_meeting, original_route_message
    original_broadcast_to_meeting = MeetingManager.broadcast_to_meeting_as_owner
    original_route_message = Program.route_message

    # Apply patches
    MeetingManager.broadcast_to_meeting_as_owner = patched_broadcast_to_meeting_as_owner
    if not stream:
        Program.route_message = patched_route_message

    pubsub = PubSub()

    # Wrap the session_log with the custom wrapper for all agents
    for agent in playbooks.program.agents:
        if hasattr(agent, "state") and hasattr(agent.state, "session_log"):
            wrapper = SessionLogWrapper(
                agent.state.session_log, pubsub, verbose, agent, stream
            )
            agent.state.session_log = wrapper

            # Add streaming methods to the agent if it's not a human agent
            if stream and hasattr(agent, "klass") and agent.klass != "human":
                agent.start_streaming_say = wrapper.start_streaming_say
                agent.stream_say_update = wrapper.stream_say_update
                agent.complete_streaming_say = wrapper.complete_streaming_say

    def log_event(event: Event):
        print(event)

    # Connect to debug adapter if requested
    if enable_debug:
        # Start debug server with stop-on-entry flag
        debug(f"Starting debug server with agents: {playbooks.program.agents}")
        await playbooks.program.start_debug_server(
            host=debug_host, port=debug_port, stop_on_entry=stop_on_entry
        )

        # If wait_for_client is True, wait for debug adapter to connect
        if wait_for_client:
            console.print(
                f"[yellow]Waiting for debug client to connect at {debug_host}:{debug_port}...[/yellow]"
            )
            await playbooks.program._debug_server.wait_for_client()
            console.print("[green]Debug client connected.[/green]")

    # Start the program
    try:
        if verbose:
            playbooks.event_bus.subscribe("*", log_event)
        await playbooks.program.run_till_exit()
    except ExecutionFinished:
        user_output.success(f"{EXECUTION_FINISHED}. Exiting...")
    except KeyboardInterrupt:
        user_output.info("Exiting...")
    except Exception as e:
        user_output.error("Execution error", details=str(e))
        raise
    finally:
        # Check for agent errors after execution using standardized error handling
        check_playbooks_health(
            playbooks,
            print_errors=True,
            log_errors=True,
            raise_on_errors=False,  # Don't raise in CLI context
            context="agent_chat_execution",
        )
        if verbose:
            playbooks.event_bus.unsubscribe("*", log_event)
        # Shutdown debug server if it was started
        if enable_debug and playbooks.program._debug_server:
            await playbooks.program.shutdown_debug_server()
        # Restore the original methods when we're done
        MessagingMixin.WaitForMessage = original_wait_for_message
        if original_broadcast_to_meeting:
            MeetingManager.broadcast_to_meeting_as_owner = original_broadcast_to_meeting
        if original_route_message:
            Program.route_message = original_route_message


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Run the agent chat application.")
    parser.add_argument(
        "program_paths",
        help="Paths to the playbook file(s) to load",
        nargs="+",
    )
    parser.add_argument(
        "-v", "--verbose", action="store_true", help="Print the session log"
    )
    parser.add_argument("--debug", action="store_true", help="Start the debug server")
    parser.add_argument(
        "--debug-host",
        default="127.0.0.1",
        help="Debug server host (default: 127.0.0.1)",
    )
    parser.add_argument(
        "--debug-port", type=int, default=7529, help="Debug server port (default: 7529)"
    )
    parser.add_argument(
        "--wait-for-client",
        action="store_true",
        help="Wait for a debug client to connect before starting execution",
    )
    parser.add_argument(
        "--skip-compilation",
        action="store_true",
        help="Skip compilation (automatically skipped for .pbasm files)",
    )
    parser.add_argument(
        "--stop-on-entry",
        action="store_true",
        help="Stop at the beginning of playbook execution",
    )
    parser.add_argument(
        "--stream",
        type=lambda x: x.lower() in ["true", "1", "yes"],
        default=True,
        help="Enable/disable streaming output (default: True). Use --stream=False for buffered output",
    )
    args = parser.parse_args()

    try:
        asyncio.run(
            main(
                args.program_paths,
                args.verbose,
                args.debug,
                args.debug_host,
                args.debug_port,
                args.wait_for_client,
                args.stop_on_entry,
                args.stream,
            )
        )
    except KeyboardInterrupt:
        print("\nGracefully shutting down...")



import logging
from typing import TYPE_CHECKING, Any, Dict, Type

from playbooks.agents.namespace_manager import AgentNamespaceManager
from playbooks.event_bus import EventBus
from playbooks.exceptions import AgentConfigurationError
from playbooks.playbook import LLMPlaybook, PythonPlaybook
from playbooks.utils.markdown_to_ast import refresh_markdown_attributes

from .ai_agent import AIAgent
from .registry import AgentClassRegistry

logger = logging.getLogger(__name__)

if TYPE_CHECKING:
    from ..program import Program


class LocalAIAgent(AIAgent):
    """
    Local AI agent that executes playbooks locally.

    This agent executes markdown and Python playbooks within the local process,
    using the existing execution infrastructure.
    """

    @classmethod
    def can_handle(cls, metadata: Dict[str, Any]) -> bool:
        """Check if this agent class can handle the given metadata.

        LocalAIAgent handles configurations without remote settings,
        or any configuration that other agents can't handle (fallback).

        Args:
            metadata: Agent metadata from playbook

        Returns:
            bool: True if this agent can handle the metadata
        """
        # LocalAIAgent handles any configuration without remote settings
        return "remote" not in metadata

    @staticmethod
    def create_class(
        klass: str,
        description: str,
        metadata: Dict[str, Any],
        h1: Dict,
        source_line_number: int,
        namespace_manager: AgentNamespaceManager = None,
    ) -> Type["LocalAIAgent"]:
        """Create and return a new local Agent class.

        Args:
            klass: Agent class name (from source)
            description: Agent description
            metadata: Agent metadata
            source_line_number: Line number in source where agent is defined

        Returns:
            Type[LocalAIAgent]: Dynamically created Agent class

        Raises:
            AgentConfigurationError: If agent class already exists
        """

        playbooks = {}

        # First create markdown playbooks to ensure their namespace functions are available
        markdown_playbooks = LLMPlaybook.create_playbooks_from_h1(h1, namespace_manager)
        playbooks.update(markdown_playbooks)

        # Then create python playbooks so they can access the markdown playbook functions
        python_playbooks = PythonPlaybook.create_playbooks_from_h1(
            h1, namespace_manager
        )
        for playbook in python_playbooks.values():
            playbook.source_file_path = h1.get("source_file_path")
            playbook.agent_name = klass
        playbooks.update(python_playbooks)

        if not playbooks:
            raise AgentConfigurationError(f"No playbooks defined for AI agent {klass}")

        # Refresh markdown attributes to ensure Python code is not sent to the LLM
        refresh_markdown_attributes(h1)

        # Extract source file path from h1 node
        source_file_path = h1.get("source_file_path")

        # Define __init__ for the new class
        def __init__(self, event_bus: EventBus, agent_id: str = None, **kwargs):
            LocalAIAgent.__init__(
                self,
                event_bus=event_bus,
                source_line_number=source_line_number,
                source_file_path=source_file_path,
                agent_id=agent_id,
                **kwargs,
            )

        # Create and return the new Agent class
        return type(
            klass,
            (LocalAIAgent,),
            {
                "__init__": __init__,
                "klass": klass,
                "description": description,
                "playbooks": playbooks,
                "metadata": metadata,
                "namespace_manager": namespace_manager,
            },
        )

    def __init__(
        self,
        event_bus: EventBus,
        source_line_number: int = None,
        source_file_path: str = None,
        agent_id: str = None,
        program: "Program" = None,
        **kwargs,
    ):
        """Initialize a new LocalAIAgent.

        Args:
            klass: The class/type of this agent.
            description: Human-readable description of the agent.
            event_bus: The event bus for publishing events.
            playbooks: Dictionary of playbooks available to this agent.
            source_line_number: The line number in the source markdown where this
                agent is defined.
            agent_id: Optional agent ID. If not provided, will generate UUID.
        """
        super().__init__(
            event_bus=event_bus,
            source_line_number=source_line_number,
            source_file_path=source_file_path,
            agent_id=agent_id,
            program=program,
            **kwargs,
        )
        # Namespace setup is now handled in AIAgent.__init__

    async def discover_playbooks(self) -> None:
        """Discover playbooks for local agent.

        For LocalAIAgent, playbooks are already provided during initialization,
        so this method is a no-op.
        """
        pass


# Register LocalAIAgent with the registry as a fallback (lowest priority)
AgentClassRegistry.register(LocalAIAgent, priority=0)



import logging
from typing import TYPE_CHECKING, Any, Dict, Type

from playbooks.event_bus import EventBus
from playbooks.exceptions import AgentConfigurationError
from playbooks.playbook import RemotePlaybook
from playbooks.transport import MCPTransport

from .ai_agent import AIAgentMeta
from .registry import AgentClassRegistry
from .remote_ai_agent import RemoteAIAgent

logger = logging.getLogger(__name__)

if TYPE_CHECKING:
    from ..program import Program


class MCPAgentMeta(AIAgentMeta):
    """Meta class for MCPAgent."""

    def should_create_instance_at_start(self) -> bool:
        """Whether to create an instance of the agent at start.

        MCP agents are always created at start in standby mode.
        """
        return True


class MCPAgent(RemoteAIAgent, metaclass=MCPAgentMeta):
    """
    MCP (Model Context Protocol) agent implementation.

    This agent connects to MCP servers and exposes their tools as playbooks.
    """

    @classmethod
    def can_handle(cls, metadata: Dict[str, Any]) -> bool:
        """Check if this agent class can handle the given metadata.

        MCPAgent handles configurations with remote.type == "mcp".

        Args:
            metadata: Agent metadata from playbook

        Returns:
            bool: True if this agent can handle the metadata
        """
        return (
            "remote" in metadata
            and isinstance(metadata["remote"], dict)
            and metadata["remote"].get("type") == "mcp"
        )

    @staticmethod
    def create_class(
        klass: str,
        description: str,
        metadata: Dict[str, Any],
        h1: Dict,
        source_line_number: int,
        namespace_manager=None,
    ) -> Type["MCPAgent"]:
        """Create an MCP agent class.

        Args:
            klass: Agent class name
            description: Agent description
            metadata: Agent metadata containing remote config
            h1: AST node for the agent definition
            source_line_number: Line number in source where agent is defined

        Returns:
            Type[MCPAgent]: Dynamically created MCP Agent class

        Raises:
            AgentConfigurationError: If agent class already exists
        """
        MCPAgent.validate(klass, metadata["remote"])

        remote_config = metadata["remote"]

        # Define __init__ for the new MCP agent class
        def __init__(
            self,
            event_bus: EventBus,
            agent_id: str = None,
            **kwargs,
        ):
            MCPAgent.__init__(
                self,
                event_bus=event_bus,
                remote_config=remote_config,
                source_line_number=source_line_number,
                agent_id=agent_id,
                **kwargs,
            )

        # Create and return the new MCP Agent class
        return type(
            klass,
            (MCPAgent,),
            {
                "__init__": __init__,
                "klass": klass,
                "description": description,
                "playbooks": {},
                "metadata": metadata,
            },
        )

    @staticmethod
    def validate(agent_name: str, remote_config: Dict[str, Any]) -> None:
        """Validate MCP agent configuration comprehensively.

        Args:
            agent_name: Name of the agent being configured
            remote_config: Remote configuration dictionary

        Raises:
            AgentConfigurationError: If configuration is invalid
        """
        # Check if URL is present
        if "url" not in remote_config:
            raise AgentConfigurationError(
                f"MCP agent '{agent_name}' requires 'url' in remote configuration"
            )

        # Validate URL format
        url = remote_config["url"]
        if not isinstance(url, str):
            raise AgentConfigurationError(
                f"MCP agent '{agent_name}' requires a valid URL string, got: {type(url).__name__}"
            )

        if not url.strip():
            raise AgentConfigurationError(
                f"MCP agent '{agent_name}' requires a valid URL string, got empty string"
            )

        # Validate transport type if specified
        transport = remote_config.get("transport")
        if transport is not None:
            valid_transports = [
                "sse",
                "stdio",
                "websocket",
                "streamable-http",
                "memory",
            ]
            if transport not in valid_transports:
                raise AgentConfigurationError(
                    f"MCP agent '{agent_name}' has invalid transport '{transport}'. "
                    f"Valid options: {', '.join(valid_transports)}"
                )

        # Validate timeout if specified
        timeout = remote_config.get("timeout")
        if timeout is not None:
            if not isinstance(timeout, (int, float)) or timeout <= 0:
                raise AgentConfigurationError(
                    f"MCP agent '{agent_name}' timeout must be a positive number, got: {timeout}"
                )

        # Validate auth configuration if specified
        auth = remote_config.get("auth")
        if auth is not None:
            if not isinstance(auth, dict):
                raise AgentConfigurationError(
                    f"MCP agent '{agent_name}' auth configuration must be a dictionary, got: {type(auth).__name__}"
                )

            # Validate auth type if specified
            auth_type = auth.get("type")
            if auth_type is not None:
                valid_auth_types = ["api_key", "bearer", "basic", "mtls"]
                if auth_type not in valid_auth_types:
                    raise AgentConfigurationError(
                        f"MCP agent '{agent_name}' has invalid auth type '{auth_type}'. "
                        f"Valid options: {', '.join(valid_auth_types)}"
                    )

                # Validate required fields for each auth type
                if auth_type == "api_key" and not auth.get("key"):
                    raise AgentConfigurationError(
                        f"MCP agent '{agent_name}' with api_key auth requires 'key' field"
                    )
                elif auth_type == "bearer" and not auth.get("token"):
                    raise AgentConfigurationError(
                        f"MCP agent '{agent_name}' with bearer auth requires 'token' field"
                    )
                elif auth_type == "basic" and (
                    not auth.get("username") or not auth.get("password")
                ):
                    raise AgentConfigurationError(
                        f"MCP agent '{agent_name}' with basic auth requires 'username' and 'password' fields"
                    )
                elif auth_type == "mtls" and (
                    not auth.get("cert") or not auth.get("key")
                ):
                    raise AgentConfigurationError(
                        f"MCP agent '{agent_name}' with mtls auth requires 'cert' and 'key' fields"
                    )

        # Validate URL scheme matches transport
        if transport == "stdio":
            # For stdio, URL should be a file path or command
            if url.startswith(("http://", "https://", "ws://", "wss://")):
                raise AgentConfigurationError(
                    f"MCP agent '{agent_name}' with stdio transport should not use HTTP/WebSocket URL"
                )
        elif transport in ["sse", "streamable-http"]:
            # For HTTP-based transports, URL should be HTTP(S)
            if not url.startswith(("http://", "https://")):
                raise AgentConfigurationError(
                    f"MCP agent '{agent_name}' with {transport} transport requires HTTP(S) URL"
                )
        elif transport == "websocket":
            # For WebSocket, URL should be ws(s)://
            if not url.startswith(("ws://", "wss://", "http://", "https://")):
                raise AgentConfigurationError(
                    f"MCP agent '{agent_name}' with websocket transport requires WebSocket or HTTP URL"
                )

    def __init__(
        self,
        event_bus: EventBus,
        remote_config: Dict[str, Any],
        source_line_number: int = None,
        agent_id: str = None,
        program: "Program" = None,
        **kwargs,
    ):
        """Initialize an MCP agent.

        Args:
            event_bus: The event bus for publishing events.
            remote_config: MCP server configuration containing:
                - url: MCP server URL or command
                - transport: Transport type (sse, stdio, etc.)
                - auth: Optional authentication config
                - timeout: Optional timeout in seconds
            source_line_number: The line number in the source markdown where this
                agent is defined.
            agent_id: Optional agent ID. If not provided, will generate UUID.
        """
        super().__init__(
            event_bus=event_bus,
            remote_config=remote_config,
            source_line_number=source_line_number,
            agent_id=agent_id,
            program=program,
            **kwargs,
        )
        self.transport = MCPTransport(remote_config)

    async def discover_playbooks(self) -> None:
        """Discover MCP tools and create RemotePlaybook instances for each."""
        if not self._connected:
            await self.connect()

        try:
            logger.debug(f"Discovering MCP tools for agent {self.klass}")
            tools = await self.transport.list_tools()

            # Clear existing playbooks
            self.playbooks.clear()

            # Create RemotePlaybook for each MCP tool
            for tool in tools:
                # Handle both dict-style and object-style tool representations
                if hasattr(tool, "name"):
                    # FastMCP Tool object
                    tool_name = tool.name
                    tool_description = getattr(
                        tool, "description", f"MCP tool: {tool.name}"
                    )

                    # Handle input schema properly
                    if hasattr(tool, "inputSchema"):
                        if hasattr(tool.inputSchema, "model_dump"):
                            input_schema = tool.inputSchema.model_dump()
                        elif hasattr(tool.inputSchema, "dict"):
                            input_schema = tool.inputSchema.dict()
                        else:
                            input_schema = tool.inputSchema
                    else:
                        input_schema = {}
                else:
                    # Dict-style tool
                    tool_name = tool.get("name")
                    tool_description = tool.get("description", f"MCP tool: {tool_name}")
                    input_schema = tool.get("inputSchema", {})

                if not tool_name:
                    logger.warning(f"MCP tool missing name: {tool}")
                    continue

                # Create execution function for this tool - fix closure issue
                def create_execute_fn(tool_name, schema):
                    async def execute_fn(*args, **kwargs):
                        # Convert positional args to kwargs if needed
                        if args and not kwargs:
                            # If only positional args, try to map them to the first parameter
                            properties = schema.get("properties", {})
                            if len(args) == 1 and len(properties) == 1:
                                param_name = list(properties.keys())[0]
                                kwargs = {param_name: args[0]}
                            else:
                                # Multiple args - create numbered parameters
                                kwargs = {f"arg_{i}": arg for i, arg in enumerate(args)}

                        result = await self.transport.call_tool(tool_name, kwargs)
                        result_str = str(result.content[0].text)
                        if result.is_error:
                            result_str = f"Error: {result_str}"
                        return result_str

                    return execute_fn

                execute_fn = create_execute_fn(tool_name, input_schema)

                # Extract parameter schema
                parameters = (
                    input_schema.get("properties", {})
                    if isinstance(input_schema, dict)
                    else {}
                )

                # Create RemotePlaybook
                playbook = RemotePlaybook(
                    name=tool_name,
                    description=tool_description,
                    agent_name=self.klass,
                    execute_fn=execute_fn,
                    parameters=parameters,
                    timeout=self.remote_config.get("timeout"),
                    metadata={"public": True},  # MCP tools are public by default
                )

                self.playbooks[tool_name] = playbook

            logger.info(
                f"Discovered {len(self.playbooks)} MCP tools for agent {self.klass}"
            )

            self.__class__.playbooks = self.playbooks
        except Exception as e:
            logger.error(
                f"Failed to discover MCP tools for agent {self.klass}: {str(e)}"
            )
            raise

    async def begin(self):
        # MCP agent does not receive messages, nor has BGN playbooks, so we do nothing
        pass


# Register MCPAgent with the registry with higher priority than LocalAIAgent
AgentClassRegistry.register(MCPAgent, priority=10)



"""
Import processor for handling !import directives in Playbooks.

This module provides functionality to process import directives in playbook files,
resolving file paths, detecting circular dependencies, and preserving indentation.
"""

import re
from pathlib import Path
from typing import Dict, List, Optional
from urllib.parse import urlparse

from .debug_logger import debug
from .exceptions import ProgramLoadError


class CircularImportError(ProgramLoadError):
    """Raised when a circular import dependency is detected."""

    def __init__(self, import_chain: List[Path]):
        self.import_chain = import_chain
        chain_str = " → ".join(str(p) for p in import_chain)
        super().__init__(f"Circular import detected: {chain_str}")


class ImportDepthError(ProgramLoadError):
    """Raised when maximum import nesting depth is exceeded."""

    def __init__(self, depth: int, max_depth: int, import_chain: List[Path]):
        self.depth = depth
        self.max_depth = max_depth
        self.import_chain = import_chain
        chain_str = " → ".join(str(p) for p in import_chain)
        super().__init__(
            f"Maximum import depth ({max_depth}) exceeded at depth {depth}. "
            f"Import chain: {chain_str}"
        )


class ImportNotFoundError(ProgramLoadError):
    """Raised when an imported file cannot be found."""

    def __init__(self, file_path: str, importing_file: Path, line_num: int):
        self.file_path = file_path
        self.importing_file = importing_file
        self.line_num = line_num
        super().__init__(
            f"Cannot import '{file_path}' - file not found\n"
            f"Location: {importing_file}, line {line_num}"
        )


class ImportProcessor:
    """Processes !import directives in playbook files."""

    # Regex pattern to match import directives
    IMPORT_PATTERN = re.compile(r"^(\s*)!import\s+(.+?)(?:\s*#.*)?$")

    # Default configuration
    DEFAULT_MAX_DEPTH = 10
    DEFAULT_MAX_FILE_SIZE = 1024 * 1024  # 1MB

    def __init__(
        self,
        base_path: Optional[Path] = None,
        max_depth: int = DEFAULT_MAX_DEPTH,
        max_file_size: int = DEFAULT_MAX_FILE_SIZE,
    ):
        """
        Initialize the import processor.

        Args:
            base_path: Base directory for resolving relative imports
            max_depth: Maximum nesting depth for imports
            max_file_size: Maximum size of imported files in bytes
        """
        self.base_path = Path(base_path) if base_path else Path.cwd()
        self.max_depth = max_depth
        self.max_file_size = max_file_size

        # Track import state during processing
        self.import_stack: List[Path] = []
        self.processed_files: Dict[Path, str] = {}

    def process_imports(self, content: str, file_path: Path, depth: int = 0) -> str:
        """
        Process all import directives in the given content.

        Args:
            content: The content containing import directives
            file_path: Path of the file being processed
            depth: Current import depth (for recursion tracking)

        Returns:
            Processed content with imports inlined

        Raises:
            CircularImportError: If circular dependency detected
            ImportDepthError: If maximum nesting depth exceeded
            ImportNotFoundError: If imported file not found
        """
        # Check depth limit
        if depth > self.max_depth:
            raise ImportDepthError(depth, self.max_depth, self.import_stack)

        # Resolve to absolute path
        file_path = file_path.resolve()

        # Check for circular imports
        if file_path in self.import_stack:
            raise CircularImportError(self.import_stack + [file_path])

        # Add to import stack
        self.import_stack.append(file_path)

        try:
            # Process the content line by line
            lines = content.split("\n")
            result_lines = []

            for line_num, line in enumerate(lines, 1):
                match = self.IMPORT_PATTERN.match(line)

                if match:
                    # Extract indentation and import path
                    indentation = match.group(1)
                    import_path = match.group(2).strip()

                    # Process the import
                    imported_lines = self._process_single_import(
                        import_path, file_path, line_num, indentation, depth
                    )
                    result_lines.extend(imported_lines)
                else:
                    # Regular line - add as-is
                    result_lines.append(line)

            return "\n".join(result_lines)

        finally:
            # Remove from import stack
            self.import_stack.pop()

    def _process_single_import(
        self,
        import_path: str,
        importing_file: Path,
        line_num: int,
        indentation: str,
        depth: int,
    ) -> List[str]:
        """
        Process a single import directive.

        Args:
            import_path: Path specified in the import directive
            importing_file: File containing the import directive
            line_num: Line number of the import directive
            indentation: Indentation to apply to imported content
            depth: Current import depth

        Returns:
            List of processed lines from the imported file

        Raises:
            ImportNotFoundError: If imported file not found
        """
        # Resolve the import path
        resolved_path = self._resolve_import_path(import_path, importing_file)

        if not resolved_path:
            raise ImportNotFoundError(import_path, importing_file, line_num)

        # Check if file is already cached
        if resolved_path in self.processed_files:
            content = self.processed_files[resolved_path]
        else:
            # Read the file content
            content = self._read_file(resolved_path, importing_file, line_num)

            # Process nested imports recursively
            if "!import" in content:
                content = self.process_imports(content, resolved_path, depth + 1)

            # Cache the processed content
            self.processed_files[resolved_path] = content

        # Apply indentation to each line
        lines = content.split("\n")
        indented_lines = [
            indentation + line if line.strip() else line for line in lines
        ]

        return indented_lines

    def _resolve_import_path(
        self, import_path: str, importing_file: Path
    ) -> Optional[Path]:
        """
        Resolve an import path to an absolute file path.

        Args:
            import_path: Path specified in import directive
            importing_file: File containing the import

        Returns:
            Resolved absolute path or None if not found
        """
        # Check if it's a URL
        parsed = urlparse(import_path)
        if parsed.scheme in ("http", "https", "file"):
            # Handle URL imports (simplified for now)
            if parsed.scheme == "file":
                return Path(parsed.path)
            else:
                # For HTTP/HTTPS, would need to download and cache
                # Not implemented in MVP
                debug(f"URL imports not yet supported: {import_path}")
                return None

        # Convert to Path object
        path = Path(import_path)

        # Try different resolution strategies
        candidates = []

        # 1. Absolute path
        if path.is_absolute():
            candidates.append(path)
        else:
            # 2. Relative to importing file's directory
            candidates.append(importing_file.parent / path)

            # 3. Relative to base path
            candidates.append(self.base_path / path)

            # 4. Relative to current working directory
            candidates.append(Path.cwd() / path)

        # Check each candidate
        for candidate in candidates:
            if candidate.exists() and candidate.is_file():
                return candidate.resolve()

        return None

    def _read_file(self, file_path: Path, importing_file: Path, line_num: int) -> str:
        """
        Read content from a file with size validation.

        Args:
            file_path: Path to read
            importing_file: File that imports this file
            line_num: Line number of import directive

        Returns:
            File content as string

        Raises:
            ImportNotFoundError: If file cannot be read
            ProgramLoadError: If file exceeds size limit
        """
        try:
            # Check file size
            file_size = file_path.stat().st_size
            if file_size > self.max_file_size:
                raise ProgramLoadError(
                    f"Imported file '{file_path}' exceeds maximum size "
                    f"({file_size} > {self.max_file_size} bytes)"
                )

            # Read the file
            with open(file_path, "r", encoding="utf-8") as f:
                return f.read()

        except (IOError, OSError) as e:
            raise ImportNotFoundError(str(file_path), importing_file, line_num) from e

    def reset(self):
        """Reset the processor state for a new processing session."""
        self.import_stack.clear()
        self.processed_files.clear()



"""
Example input with both metadata and description --
metadata:
  framework: GAAP
  specialization:
    - accounting
    - tax
  author: John Doe
---
This is an accountant agent that can help with accounting tasks.

Example input with only description --
This is an accountant agent that can help with accounting tasks.

Example input with only metadata --
metadata:
  framework: GAAP
  specialization:
    - accounting
    - tax
  author: John Doe

Also, input can be empty.
"""

import yaml


def parse_metadata_and_description(input: str) -> tuple[dict, str]:
    """Parse the input into a metadata and description."""
    if not input or not input.strip():
        return {}, ""

    if input.startswith("metadata:"):
        parts = input.split("---", maxsplit=1)
        if len(parts) == 1:
            metadata_text = parts[0]
            description_text = ""
        else:
            metadata_text = parts[0]
            description_text = parts[1]
        metadata = yaml.safe_load(metadata_text)["metadata"] or {}
        description = description_text.strip()
    else:
        metadata = {}
        description = input.strip()

    return metadata, description



def get_playbooks_version() -> str:
    """Get the version of the playbooks package."""
    try:
        from importlib.metadata import version

        return version("playbooks")
    except ImportError:
        # Fallback for Python < 3.8
        try:
            from importlib_metadata import version

            return version("playbooks")
        except ImportError:
            return "unknown"
    except Exception:
        return "unknown"



import copy
import tempfile
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any, Dict, List

from ..call_stack import CallStackFrame, InstructionPointer
from ..constants import EXECUTION_FINISHED, HUMAN_AGENT_KLASS
from ..debug_logger import debug
from ..enums import StartupMode
from ..event_bus import EventBus
from ..exceptions import ExecutionFinished
from ..execution_state import ExecutionState
from ..llm_messages import (
    ExecutionResultLLMMessage,
    FileLoadLLMMessage,
    MeetingLLMMessage,
)
from ..meetings import MeetingManager
from ..playbook import LLMPlaybook, Playbook, PythonPlaybook, RemotePlaybook
from ..playbook_call import PlaybookCall, PlaybookCallResult
from ..utils.expression_engine import (
    ExpressionContext,
)
from ..utils.langfuse_helper import LangfuseHelper
from ..utils.misc import copy_func
from ..utils.spec_utils import SpecUtils
from .base_agent import BaseAgent, BaseAgentMeta
from .namespace_manager import AgentNamespaceManager

if TYPE_CHECKING:
    from ..program import Program


class AIAgentMeta(BaseAgentMeta):
    """Meta class for AIAgent."""

    def __new__(cls, name, bases, attrs):
        cls = super().__new__(cls, name, bases, attrs)
        cls.validate_metadata()
        return cls

    @property
    def startup_mode(self) -> StartupMode:
        """Get the startup mode for this agent."""
        return getattr(self, "metadata", {}).get("startup_mode", StartupMode.DEFAULT)

    def validate_metadata(self):
        """Validate the metadata for this agent."""
        if self.startup_mode not in [StartupMode.DEFAULT, StartupMode.STANDBY]:
            raise ValueError(f"Invalid startup mode: {self.startup_mode}")

    def should_create_instance_at_start(self) -> bool:
        """Whether to create an instance of the agent at start.

        Override in subclasses to control whether to create an instance at start.
        """
        # If there is any playbook with a BGN trigger, return True
        for playbook in self.playbooks.values():
            if playbook.triggers:
                for trigger in playbook.triggers.triggers:
                    if trigger.is_begin:
                        return True

        # This agent does not have any BGN playbook
        # Check if it should be created in standby mode
        if self.startup_mode == StartupMode.STANDBY:
            return True

        return False


class AIAgent(BaseAgent, ABC, metaclass=AIAgentMeta):
    """
    Abstract base class for AI agents.

    An Agent represents an AI entity capable of processing messages through playbooks
    using a main execution thread. This class defines the interface that all AI agent
    implementations must adhere to.

    Attributes:
        klass: The class/type of this agent.
        description: Human-readable description of the agent.
        playbooks: Dictionary of playbooks available to this agent.
    """

    def __init__(
        self,
        event_bus: EventBus,
        source_line_number: int = None,
        source_file_path: str = None,
        agent_id: str = None,
        program: "Program" = None,
        **kwargs,
    ):
        """Initialize a new AIAgent.

        Args:
            klass: The class/type of this agent.
            description: Human-readable description of the agent.
            event_bus: The event bus for publishing events.
            playbooks: Dictionary of playbooks available to this agent.
            source_line_number: The line number in the source markdown where this
                agent is defined.
            agent_id: Optional agent ID. If not provided, will generate UUID.
        """
        super().__init__(
            agent_id=agent_id,
            program=program,
            source_line_number=source_line_number,
            source_file_path=source_file_path,
            **kwargs,
        )
        self.playbooks: Dict[str, Playbook] = self.deep_copy_playbooks(
            self.__class__.playbooks or {}
        )
        # Create instance-specific namespace with playbook wrappers
        self._setup_isolated_namespace()

        # Initialize meeting manager
        self.meeting_manager = MeetingManager(agent=self)

        self.meeting_manager.ensure_meeting_playbook_kwargs(self.playbooks)

        self.state = ExecutionState(event_bus, self.klass, self.id)
        self.source_line_number = source_line_number
        self.public_json = None

        # Track background tasks for cleanup
        # self._background_tasks = []

        # Create playbook to run BGN playbooks
        self.bgn_playbook_name = None
        self.create_begin_playbook()

    def deep_copy_playbooks(self, playbooks):
        """Deep copy the playbooks."""
        playbooks_copy = copy.deepcopy(playbooks)
        for playbook in playbooks_copy.values():
            if playbook.func:
                playbook.func = copy_func(playbook.func)

        return playbooks_copy

    def _setup_isolated_namespace(self):
        """Create isolated namespace with instance-specific agent reference and playbook wrappers."""
        # Create isolated namespace for this instance
        # Preserve class-level namespace if it exists (contains imports from Python code blocks)
        if (
            hasattr(self.__class__, "namespace_manager")
            and self.__class__.namespace_manager
        ):
            # Copy the class namespace to preserve imports and module-level variables
            self.namespace_manager = AgentNamespaceManager(
                namespace=self.__class__.namespace_manager.namespace.copy()
            )
        else:
            self.namespace_manager = AgentNamespaceManager()
        self.namespace_manager.namespace["agent"] = self

        # Set up cross-playbook wrapper functions and bind agent-specific functions
        for playbook_name, playbook in self.playbooks.items():
            # Create cross-playbook wrapper function
            call_through = playbook.create_namespace_function(self)
            self.namespace_manager.namespace[playbook_name] = call_through
            playbook.agent_name = str(self)

        for playbook_name, playbook in self.playbooks.items():
            if (
                hasattr(playbook, "create_agent_specific_function")
                and not playbook.func
            ):
                playbook.func = playbook.create_agent_specific_function(self)
            else:
                playbook.func = copy_func(
                    playbook.func,
                    globals={
                        **playbook.func.__globals__,
                        **self.namespace_manager.namespace,
                    },
                )

    def create_agent_wrapper(self, agent, func):
        """Create an agent-specific wrapper that bypasses globals lookup."""

        async def agent_specific_wrapper(*args, _agent=agent, **kwargs):
            return await func(*args, **kwargs)

        return agent_specific_wrapper

    @abstractmethod
    async def discover_playbooks(self) -> None:
        """Discover and load playbooks for this agent.

        This method should populate the self.playbooks dictionary with
        available playbooks for this agent.
        """
        pass

    @property
    def startup_mode(self) -> StartupMode:
        """Get the startup mode for this agent."""
        return self.__class__.startup_mode

    @property
    def other_agents(self) -> List["AIAgent"]:
        """Get list of other AI agents in the system.

        Returns:
            List of other agent instances
        """
        if (
            not self.program
            or not hasattr(self.program, "agents")
            or not self.program.agents
        ):
            return []

        return list(
            filter(lambda x: isinstance(x, AIAgent) and x != self, self.program.agents)
        )

    def event_agents_changed(self):
        self.state.agents = [str(agent) for agent in self.program.agents]

    def get_available_playbooks(self) -> List[str]:
        """Get a list of available playbook names.

        Returns:
            List of playbook names available to this agent
        """
        return list(self.playbooks.keys())

    def create_begin_playbook(self):
        begin_playbooks = {}
        for playbook in self.playbooks.values():
            if hasattr(playbook, "triggers") and playbook.triggers:
                for trigger in playbook.triggers.triggers:
                    if trigger.is_begin:
                        begin_playbooks[playbook.name] = playbook

        # If there are multiple BGN playbooks, create a new playbook that calls them in order
        self.bgn_playbook_name = "Begin__"
        while self.bgn_playbook_name in self.playbooks:
            self.bgn_playbook_name = "_" + self.bgn_playbook_name
        code_block = f"""
@playbook
async def {self.bgn_playbook_name}() -> None:
    # Main loop for agent "{self.klass}"
    # Auto-generated by Playbooks AI runtime
    # 
    # Calls any playbooks that should be executed when the program starts, followed by a loop that waits for messages and processes them.
    agent.state.variables["$_busy"] = True
{"\n".join(["    await " + playbook.name + "()" for playbook in begin_playbooks.values()])}

    agent.state.variables["$_busy"] = False
    if agent.program and agent.program.execution_finished:
        return
    
    # Enter a message processing event loop
    await MessageProcessingEventLoop()
"""

        # Save to tmp file
        prefix = f"{self.klass}_{self.bgn_playbook_name}"
        file_path = None
        with tempfile.NamedTemporaryFile(
            mode="w", prefix=prefix, suffix=".pb", delete=False
        ) as f:
            f.write(code_block)
            f.flush()
            file_path = f.name

        # debug("BGN Playbook Code Block: " + code_block)
        new_playbook = PythonPlaybook.create_playbooks_from_code_block(
            code_block,
            self.namespace_manager,
            file_path,
            1,
        )
        for playbook in new_playbook.values():
            playbook.source_file_path = file_path
            playbook.agent_name = str(self)
        self.playbooks.update(new_playbook)

    async def begin(self):
        await self.execute_playbook(self.bgn_playbook_name)
        return

    async def cleanup(self):
        """Cancel all background tasks and clean up resources."""
        # Only cleanup if execution is truly finished
        if not (self.program and self.program.execution_finished):
            return

    def parse_instruction_pointer(self, step_id: str) -> InstructionPointer:
        """Parse a step string into an InstructionPointer.

        Args:
            step: Step string to parse

        Returns:
            InstructionPointer: Parsed instruction pointer
        """
        # Extract the step number from the step string
        playbook_name = step_id.split(":")[0]
        step_number = step_id.split(":")[1]
        playbook = self.playbooks.get(playbook_name)

        # Ignore trigger and note step, e.g. `PB:T1`, `PB:N1`
        if playbook and step_number[0] not in ["T", "N"] and playbook.steps:
            line = playbook.steps.get_step(step_number)
            if line:
                return InstructionPointer(
                    playbook=playbook_name,
                    line_number=step_number,
                    source_line_number=line.source_line_number,
                    step=line,
                    source_file_path=line.source_file_path,
                )
        return InstructionPointer(
            playbook=playbook_name,
            line_number=step_number,
            source_line_number=0,
            step=None,
            source_file_path=None,
        )

    def trigger_instructions(
        self,
        with_namespace: bool = False,
        public_only: bool = False,
        skip_bgn: bool = True,
    ) -> List[str]:
        """Get trigger instructions for this agent's playbooks.

        Args:
            with_namespace: Whether to include namespace in instructions
            public_only: Whether to only include public playbooks
            skip_bgn: Whether to skip BGN trigger instructions

        Returns:
            List of trigger instruction strings
        """
        instructions = []
        for playbook in self.playbooks.values():
            if public_only and not playbook.public:
                continue

            namespace = self.klass if with_namespace else None
            playbook_instructions = playbook.trigger_instructions(namespace, skip_bgn)
            instructions.extend(playbook_instructions)
        return instructions

    def all_trigger_instructions(self) -> List[str]:
        """Get all trigger instructions including from other agents.

        Returns:
            List of all trigger instruction strings
        """
        instructions = self.trigger_instructions(with_namespace=False)
        for agent in self.other_agents:
            instructions.extend(agent.trigger_instructions(with_namespace=True))
        return instructions

    @classmethod
    def get_compact_information(cls) -> str:
        info_parts = []
        info_parts.append(f"# {cls.klass}")
        if cls.description:
            info_parts.append(f"{cls.description}")

        if cls.playbooks:
            for playbook in cls.playbooks.values():
                if not playbook.hidden:
                    info_parts.append(f"## {playbook.signature}")
                    if playbook.description:
                        info_parts.append(
                            playbook.description[:100]
                            + ("..." if len(playbook.description) > 100 else "")
                        )
                    info_parts.append("\n")

        return "\n".join(info_parts)

    @classmethod
    def get_public_information(cls) -> str:
        """Get public information about an agent klass

        Returns:
            String containing public agent information
        """
        info_parts = []
        info_parts.append(f"# {cls.klass}")
        if cls.description:
            info_parts.append(f"{cls.description}")

        if cls.playbooks:
            for playbook in cls.playbooks.values():
                if playbook.public:
                    info_parts.append(f"## {cls.klass}.{playbook.name}")
                    info_parts.append(playbook.description)

        return "\n".join(info_parts)

    def other_agent_klasses_information(self) -> List[str]:
        """Get information about other registered agents.

        Returns:
            List of information strings for other agents
        """
        return [
            agent_klass.get_public_information()
            for agent_klass in self.program.agent_klasses.values()
            if agent_klass.klass != self.klass
        ]

    def resolve_target(self, target: str = None, allow_fallback: bool = True) -> str:
        """Resolve a target specification to an agent ID.

        Args:
            target: Target specification (agent ID, agent type, "human", etc.)
            allow_fallback: Whether to use fallback logic when target is None

        Returns:
            Resolved target agent ID, or None if no fallback allowed and target not found
        """
        if target is not None:
            target = target.strip()

            # Handle human aliases
            if target.lower() in ["human", "user"]:
                return "human"

            # Handle meeting targets (Phase 5)
            if target == "meeting":
                # Map "meeting" to current meeting context
                if meeting_id := self.state.get_current_meeting():
                    return f"meeting {meeting_id}"
                return None  # No current meeting

            if SpecUtils.is_meeting_spec(target):
                return target  # Return as-is for now

            # Handle agent ID targets
            if SpecUtils.is_agent_spec(target):
                agent_id = SpecUtils.extract_agent_id(target)
                return agent_id

            # Check if target is a numeric agent ID
            if target.isdigit():
                return target

            # Handle special YLD targets
            if target == "last_non_human_agent":
                if (
                    self.state.last_message_target
                    and self.state.last_message_target != "human"
                ):
                    return self.state.last_message_target
                return None  # No fallback for this case

            # Handle agent type - find first agent of this type
            for agent in self.other_agents:
                if agent.klass == target:
                    return agent.id

            # If not found, check if Human agent exists with this type name
            if target == HUMAN_AGENT_KLASS:
                return "human"

            # Target not found - fallback to human if allowed
            return "human" if allow_fallback else None

        # No target specified - use fallback logic if allowed
        if not allow_fallback:
            return None

        # Fallback logic: current context → last 1:1 target → Human
        # Check current meeting context first
        if meeting_id := self.state.get_current_meeting():
            return f"meeting {meeting_id}"

        # Check last 1:1 target
        if self.state.last_message_target:
            return self.state.last_message_target

        # Default to Human
        return "human"

    @property
    def public_playbooks(self) -> List[Playbook]:
        """Get list of public playbooks with their information.

        Returns:
            List of dictionaries containing public playbook information
        """
        public_playbooks = []
        for playbook in self.playbooks.values():
            if playbook.public:
                public_playbooks.append(playbook)
        return public_playbooks

    def _build_input_log(self, playbook: Playbook, call: PlaybookCall) -> str:
        """Build the input log string for Langfuse tracing.

        Args:
            playbook: The playbook being executed
            call: The playbook call information

        Returns:
            A string containing the input log data
        """
        log_parts = []
        log_parts.append(str(self.state.call_stack))
        log_parts.append(str(self.state.variables))
        log_parts.append("Session log: \n" + str(self.state.session_log))

        if isinstance(playbook, LLMPlaybook):
            log_parts.append(playbook.markdown)
        elif isinstance(playbook, PythonPlaybook):
            log_parts.append(playbook.code or f"Python function: {playbook.name}")
        elif isinstance(playbook, RemotePlaybook):
            log_parts.append(playbook.__repr__())

        log_parts.append(str(call))

        return "\n\n".join(log_parts)

    async def _pre_execute(
        self, playbook_name: str, args: List[Any], kwargs: Dict[str, Any]
    ) -> tuple:
        call = PlaybookCall(playbook_name, args, kwargs)
        playbook = self.playbooks.get(playbook_name)

        trace_str = str(self) + "." + call.to_log_full()

        if playbook:
            # Set up tracing
            if isinstance(playbook, LLMPlaybook):
                trace_str = f"Markdown: {trace_str}"
            elif isinstance(playbook, PythonPlaybook):
                trace_str = f"Python: {trace_str}"
            elif isinstance(playbook, RemotePlaybook):
                trace_str = f"Remote: {trace_str}"
        else:
            trace_str = f"External: {trace_str}"

        if self.state.call_stack.peek() is not None:
            langfuse_span = self.state.call_stack.peek().langfuse_span.span(
                name=trace_str
            )
        else:
            langfuse_span = LangfuseHelper.instance().trace(name=trace_str)

        if playbook:
            input_log = self._build_input_log(playbook, call)
            langfuse_span.update(input=input_log)
        else:
            langfuse_span.update(input=trace_str)

        # Add the call to the call stack
        if playbook:
            # Get first step line number if available (for LLMPlaybook)
            first_step_line_number = (
                getattr(playbook, "first_step_line_number", None) or 0
            )
        else:
            first_step_line_number = 0

        # Check if this is a meeting playbook and get meeting context
        is_meeting = False
        meeting_id = None
        if playbook and playbook.meeting:
            is_meeting = True
            # Try to get meeting ID from kwargs or current context
            meeting_id = kwargs.get("meeting_id") or self.state.get_current_meeting()

        source_file_path = (
            playbook.source_file_path
            if playbook and hasattr(playbook, "source_file_path")
            else None
        )
        source_file_path = source_file_path or "[unknown]"
        call_stack_frame = CallStackFrame(
            InstructionPointer(
                playbook=call.playbook_klass,
                line_number="01",
                source_line_number=first_step_line_number,
                source_file_path=source_file_path,
            ),
            langfuse_span=langfuse_span,
            is_meeting=is_meeting,
            meeting_id=meeting_id,
        )
        self.state.call_stack.push(call_stack_frame)
        self.state.session_log.append(call)

        self.state.variables.update({"$__": None})

        return playbook, call, langfuse_span

    async def execute_playbook(
        self, playbook_name: str, args: List[Any] = [], kwargs: Dict[str, Any] = {}
    ) -> Any:
        if self.program and self.program.execution_finished:
            return EXECUTION_FINISHED

        playbook, call, langfuse_span = await self._pre_execute(
            playbook_name, args, kwargs
        )

        # Replace variable names with actual values
        context = ExpressionContext(agent=self, state=self.state, call=call)

        # Resolve args
        for i, arg in enumerate(args):
            if isinstance(arg, str) and arg.startswith("$"):
                try:
                    args[i] = context.evaluate_expression(arg)
                except Exception:
                    # If resolution fails, keep the original value
                    pass

        # Resolve kwargs
        for key, value in kwargs.items():
            if isinstance(value, str) and value.startswith("$"):
                try:
                    kwargs[key] = context.evaluate_expression(value)
                except Exception:
                    # If resolution fails, keep the original value
                    pass

        try:
            # Handle meeting playbook initialization (only for new meetings, not when joining existing ones)
            debug(f"{str(self)}: Handling meeting playbook execution: {playbook_name}")
            debug(
                f"{str(self)}: Current meeting from call stack: {self.meeting_manager.get_current_meeting_from_call_stack()}"
            )
            if (
                playbook
                and playbook.meeting
                and not self.meeting_manager.get_current_meeting_from_call_stack()
            ):
                meeting = await self.meeting_manager.create_meeting(
                    playbook_name, kwargs
                )

                if self.program and self.program.execution_finished:
                    return EXECUTION_FINISHED

                # Wait for required attendees to join before proceeding (if any besides requester)
                await self.meeting_manager._wait_for_required_attendees(meeting)

                message = f"Meeting {meeting.id} ready to proceed - all required attendees present"
                self.state.session_log.append(message)

                meeting_msg = MeetingLLMMessage(message, meeting_id=meeting.id)
                self.state.call_stack.add_llm_message(meeting_msg)
        except TimeoutError as e:
            error_msg = f"Meeting initialization failed: {str(e)}"
            await self._post_execute(call, error_msg, langfuse_span)
            return error_msg

        # Execute local playbook in this agent
        if playbook:
            try:
                if self.program and self.program.execution_finished:
                    return EXECUTION_FINISHED

                result = await playbook.execute(*args, **kwargs)
                await self._post_execute(call, result, langfuse_span)
                return result
            except ExecutionFinished as e:
                debug("Execution finished, exiting", agent=str(self))
                self.program.set_execution_finished(reason="normal", exit_code=0)
                message = str(e)
                await self._post_execute(call, message, langfuse_span)
                return message
            except Exception as e:
                message = f"Error: {str(e)}"
                await self._post_execute(call, message, langfuse_span)
                raise
        else:
            # Handle cross-agent playbook calls (AgentName.PlaybookName format)
            if "." in playbook_name:
                agent_name, actual_playbook_name = playbook_name.split(".", 1)
                target_agent = list(
                    filter(lambda x: x.klass == agent_name, self.program.agents)
                )
                if target_agent:
                    target_agent = target_agent[0]

                if (
                    target_agent
                    and actual_playbook_name in target_agent.playbooks
                    and target_agent.playbooks[actual_playbook_name].public
                ):
                    result = await target_agent.execute_playbook(
                        actual_playbook_name, args, kwargs
                    )
                    await self._post_execute(call, result, langfuse_span)
                    return result

            # Try to execute playbook in other agents (fallback)
            for agent in self.other_agents:
                if (
                    playbook_name in agent.playbooks
                    and agent.playbooks[playbook_name].public
                ):
                    result = await agent.execute_playbook(playbook_name, args, kwargs)
                    await self._post_execute(call, result, langfuse_span)
                    return result

            # Playbook not found
            error_msg = f"Playbook '{playbook_name}' not found in agent '{self.klass}' or any registered agents"
            await self._post_execute(call, error_msg, langfuse_span)
            return error_msg

    async def _post_execute(
        self, call: PlaybookCall, result: Any, langfuse_span: Any
    ) -> None:
        execution_summary = self.state.variables.variables["$__"].value
        call_result = PlaybookCallResult(call, result, execution_summary)
        self.state.session_log.append(call_result)

        self.state.call_stack.pop()

        result_msg = ExecutionResultLLMMessage(
            call_result.to_log_full(), playbook_name=call.playbook_klass, success=True
        )
        self.state.call_stack.add_llm_message(result_msg)

        langfuse_span.update(output=result)

    def __str__(self):
        if self.kwargs:
            kwargs_msg = ", ".join([f"{k}:{v}" for k, v in self.kwargs.items()])
            return f"{self.klass}(agent {self.id}, {kwargs_msg})"
        else:
            return f"{self.klass}(agent {self.id})"

    @property
    def name(self):
        if self.kwargs and "name" in self.kwargs:
            return self.kwargs["name"]
        else:
            return self.klass

    def __repr__(self):
        return f"{self.klass}(agent {self.id})"

    async def load_file(
        self, file_path: str, inline: bool = False, silent: bool = False
    ) -> str:
        with open(file_path, "r") as file:
            content = file.read()
        if inline:
            return content
        else:
            # Safely get the caller frame (second from top)
            if len(self.state.call_stack.frames) >= 2:
                caller_frame = self.state.call_stack.frames[-2]

                if silent:
                    file_msg = FileLoadLLMMessage(content, file_path=file_path)
                    caller_frame.add_llm_message(file_msg)
                    return ""
                else:
                    file_msg = FileLoadLLMMessage(
                        f"Contents of file {file_path}:\n\n{content}",
                        file_path=file_path,
                    )
                    caller_frame.add_llm_message(file_msg)

                    return f"Loaded file {file_path}"
            else:
                # Not enough frames in call stack, just return the content
                return f"Loaded file {file_path}"
